from typing import Union, Optional
from dataclasses import dataclass, field
from box import Box
import os
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box

logger = logging_tools.set_logging()


@dataclass
class CziAddMetaData:
    """
    A class to add metadata from CZI image data.
    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): The source of the CZI image data.
        experiment (Optional[Box]): The experiment metadata. Initialized to None.
        hardwaresetting (Optional[Box]): The hardware setting metadata. Initialized to None.
        customattributes (Optional[Box]): The custom attributes metadata. Initialized to None.
        displaysetting (Optional[Box]): The display setting metadata. Initialized to None.
        layers (Optional[Box]): The layers metadata. Initialized to None.
        verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods:
        __post_init__(): Reads additional metadata from the CZI image data and initializes the attributes.
    """

    czisource: Union[str, os.PathLike[str], Box]
    experiment: Optional[Box] = field(init=False, default=None)
    hardwaresetting: Optional[Box] = field(init=False, default=None)
    customattributes: Optional[Box] = field(init=False, default=None)
    displaysetting: Optional[Box] = field(init=False, default=None)
    layers: Optional[Box] = field(init=False, default=None)
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading additional Metedata from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        if czi_box.has_experiment:
            self.experiment = czi_box.ImageDocument.Metadata.Experiment
        else:
            if self.verbose:
                logger.info("No Experiment information found.")

        if czi_box.has_hardware:
            self.hardwaresetting = czi_box.ImageDocument.Metadata.HardwareSetting
        else:
            if self.verbose:
                logger.info("No HardwareSetting information found.")

        if czi_box.has_customattr:
            self.customattributes = czi_box.ImageDocument.Metadata.CustomAttributes
        else:
            if self.verbose:
                logger.info("No CustomAttributes information found.")

        if czi_box.has_disp:
            self.displaysetting = czi_box.ImageDocument.Metadata.DisplaySetting
        else:
            if self.verbose:
                logger.info("No DisplaySetting information found.")

        if czi_box.has_layers:
            self.layers = czi_box.ImageDocument.Metadata.Layers
        else:
            if self.verbose:
                logger.info("No Layers information found.")



# -*- coding: utf-8 -*-

#################################################################
# File        : misc.py
# Author      : sebi06
#
# Disclaimer: The code is purely experimental. Feel free to
# use it at your own risk.
#
#################################################################

import os
import zarr
import pandas as pd
import dask.array as da
import numpy as np
import time
from pathlib import Path
import dateutil.parser as dt
from tqdm.contrib.itertools import product
from typing import Dict, Tuple, Any, Union, Annotated
import validators
from aicspylibczi import CziFile
from czitools.metadata_tools.helper import ValueRange
from czitools.utils import logging_tools
import requests
import time
import tracemalloc

logger = logging_tools.set_logging()


def slicedim(array: Union[np.ndarray, da.Array, zarr.Array], dimindex: int, posdim: int) -> np.ndarray:
    """Slice out a specific dimension without (!) dropping the dimension
    of the array to conserve the dimorder string
    This works for Numpy.Array, Dask and ZARR.

    Example:
        array.shape = (1, 3, 2, 5, 170, 240) and dim_order is STCZYX.
        Index for C inside array is 2.
        Task: Cut out the fist channel = 0.
        channel = slicedim(array, 0, 2).
        The resulting channel.shape = (1, 3, 1, 5, 170, 240)

    Args:
        array: Input array.
        dimindex: Index of the slice dimension to be kept.
        posdim: Position of the dimension to be sliced.

    Returns:
        Sliced array.

    """

    idl_all = [slice(None, None, None)] * (len(array.shape) - 2)
    idl_all[posdim] = slice(dimindex, dimindex + 1, None)
    array_sliced = array[tuple(idl_all)]

    return array_sliced


def calc_scaling(
    data: Union[np.ndarray, da.array],
    corr_min: float = 1.0,
    offset_min: int = 0,
    corr_max: float = 0.85,
    offset_max: int = 0,
) -> Tuple[int, int]:
    """Calculate the scaling for better display

    Args:
        data (Union[numpy.ndarray, dask.array, zarr.array]): Array to calculate scaling.
        corr_min (float, optional): Correction factor for minimum value. Defaults to 1.0.
        offset_min (int, optional): Offset for minimum value. Defaults to 0.
        corr_max (float, optional): Correction factor for maximum value. Defaults to 0.85.
        offset_max (int, optional): Offset for maximum value. Defaults to 0.

    Returns:
        Tuple[int, int]: Tuple with minimum value and maximum value.

    Raises:
        ValueError: If the data array is not of type numpy.ndarray, dask.array or zarr.array.

    """

    start = time.time()

    # get min-max values for initial scaling
    if isinstance(data, zarr.Array):
        minvalue, maxvalue = np.min(data), np.max(data)
    elif isinstance(data, da.Array):
        # compute only once since this is faster
        minvalue, maxvalue = da.compute(data.min(), data.max())
    else:
        minvalue, maxvalue = np.min(data), np.max(data)

    end = time.time()

    minvalue = int(np.round((minvalue + offset_min) * corr_min, 0))
    maxvalue = int(np.round((maxvalue + offset_max) * corr_max, 0))

    logger.info(f"Scaling: {minvalue}, {maxvalue}")
    logger.info(f"Calculation of Min-Max [s] : {end - start}")

    return minvalue, maxvalue


def md2dataframe(md_dict: Dict, paramcol: str = "Parameter", keycol: str = "Value") -> pd.DataFrame:
    """Converts the given metadata_tools dictionary to a Pandas DataFrame.

    Args:
        md_dict (dict): A dictionary containing metadata_tools.
        paramcol (str, optional): The name of the column for metadata_tools parameters. Defaults to "Parameter".
        keycol (str, optional): The name of the column for metadata_tools values. Defaults to "Value".

    Returns:
        pd.DataFrame: A Pandas DataFrame containing all the metadata_tools.
    """
    mdframe = pd.DataFrame(columns=[paramcol, keycol])

    for k in md_dict.keys():
        d = {"Parameter": k, "Value": md_dict[k]}
        df = pd.DataFrame([d], index=[0])
        mdframe = pd.concat([mdframe, df], ignore_index=True)

    return mdframe


def sort_dict_by_key(unsorted_dict: Dict) -> Dict:
    """Sort a dictionary by key names

    Args:
        unsorted_dict: the unsorted dictionary where the keys should be sorted

    Returns:
        Dictionary with keys sorted by name
    """

    sorted_keys = sorted(unsorted_dict.keys(), key=lambda x: x.lower())
    sorted_dict = {}
    for key in sorted_keys:
        sorted_dict.update({key: unsorted_dict[key]})

    return sorted_dict


def addzeros(number: int) -> str:
    """Convert a number into a string and add leading zeros.
    Typically used to construct filenames with equal lengths.

    :param number: the number
    :type number: int
    :return: zerostring - string with leading zeros
    :rtype: str
    """

    zerostring = None

    if number < 10:
        zerostring = "0000" + str(number)
    if 10 <= number < 100:
        zerostring = "000" + str(number)
    if 100 <= number < 1000:
        zerostring = "00" + str(number)
    if 1000 <= number < 10000:
        zerostring = "0" + str(number)

    return zerostring


def get_fname_woext(filepath: Union[str, os.PathLike[str]]) -> str:
    """
    Extracts the filename without its extension from a given file path.
    It also works for extensions like myfile.abc.xyz
    The output will be: myfile

    Args:
        filepath (Union[str, os.PathLike[str]]): The path to the file.
    Returns:
        str: The filename without its extension.
    """

    # create empty string
    real_extension = ""

    # get all part of the file extension
    sufs = Path(filepath).suffixes
    for s in sufs:
        real_extension = real_extension + s

    # remove real extension from filepath
    filepath_woext = filepath.replace(real_extension, "")

    return filepath_woext


def check_dimsize(mdata_entry: Union[Any, None], set2value: Any = 1) -> Union[Any, None]:
    """Check the entries for None.

    Args:
        mdata_entry: The entry to be checked.
        set2value: The value to replace None.

    Returns:
        Union of Any and None
    """

    if mdata_entry is None:
        return set2value
    if mdata_entry is not None:
        return mdata_entry


def get_planetable(
    czifile: Union[str, os.PathLike[str]],
    norm_time: bool = True,
    save_table: bool = False,
    table_separator: str = ";",
    table_index: bool = True,
    **kwargs: dict[str, int],
) -> pd.DataFrame:
    """Get the planetable from the individual subblocks.

    Args:
        czifile: The source for the CZI image file.
        norm_time: Whether to normalize the timestamps. Defaults to True.
        save_table: Whether to save the planetable to a CSV file. Defaults to False.
        table_separator: The separator for the CSV file. Defaults to ";".
        table_index: Whether to include the index in the CSV file. Defaults to True.
        **kwargs: Optional arguments for filtering the planetable. These can include:
            - scene: Scene index to filter the data.
            - tile: Tile index to filter the data.
            - time: Time index to filter the data.
            - channel: Channel index to filter the data.
            - zplane: Z index to filter the data.

    Returns:
        pd.DataFrame: The planetable as a Pandas DataFrame.
    """

    if isinstance(czifile, Path):
        # convert to string
        czifile = str(czifile)

    if validators.url(czifile):
        logger.warning("Reading PlaneTable from CZI via a link is not supported yet.")
        return None, None

    # initialize the plane table
    df_czi = pd.DataFrame(
        columns=[
            "Subblock",
            "S",
            "M",
            "T",
            "C",
            "Z",
            "X[micron]",
            "Y[micron]",
            "Z[micron]",
            "Time[s]",
            "xstart",
            "ystart",
            "width",
            "height",
        ]
    )

    # define subblock counter
    sbcount = -1

    aicsczi = CziFile(czifile)
    dims = aicsczi.get_dims_shape()

    has_s = False
    has_m = False
    has_t = False
    has_c = False
    has_z = False

    if "S" in dims[0].keys():
        size_s = dims[0]["S"][1]
        has_s = True
    else:
        size_s = 1

    if "M" in dims[0].keys():
        size_m = dims[0]["M"][1]
        has_m = True
    else:
        size_m = 1

    if "T" in dims[0].keys():
        size_t = dims[0]["T"][1]
        has_t = True
    else:
        size_t = 1

    if "C" in dims[0].keys():
        size_c = dims[0]["C"][1]
        has_c = True
    else:
        size_c = 1

    if "Z" in dims[0].keys():
        size_z = dims[0]["Z"][1]
        has_z = True
    else:
        size_z = 1

    # cast data types
    df_czi = df_czi.astype(
        {
            "Subblock": "int32",
            "S": "int32",
            "M": "int32",
            "T": "int32",
            "C": "int32",
            "Z": "int32",
            "X[micron]": "float",
            "Y[micron]": "float",
            "Z[micron]": "float",
            "Time[s]": "float",
            "xstart": "int32",
            "ystart": "int32",
            "width": "int32",
            "height": "int32",
        },
        copy=False,
        errors="ignore",
    )

    # check for scenes argument
    if "scene" in kwargs:
        s_start = kwargs["scene"]
        s_end = kwargs["scene"] + 1
    else:
        s_start = 0
        s_end = size_s

    # check for tile argument
    if "tile" in kwargs:
        m_start = kwargs["tile"]
        m_end = kwargs["tile"] + 1
    else:
        m_start = 0
        m_end = size_m

    # check for time argument
    if "time" in kwargs:
        t_start = kwargs["time"]
        t_end = kwargs["time"] + 1
    else:
        t_start = 0
        t_end = size_t

    # check for channel argument
    if "channel" in kwargs:
        c_start = kwargs["channel"]
        c_end = kwargs["channel"] + 1
    else:
        c_start = 0
        c_end = size_c

    # check for zplane argument
    if "zplane" in kwargs:
        z_start = kwargs["zplane"]
        z_end = kwargs["zplane"] + 1
    else:
        z_start = 0
        z_end = size_z

    for s, m, t, c, z in product(
        enumerate(range(s_start, s_end)),
        enumerate(range(m_start, m_end)),
        enumerate(range(t_start, t_end)),
        enumerate(range(c_start, c_end)),
        enumerate(range(z_start, z_end)),
        desc="Reading sublocks planes",
        unit=" 2Dplanes",
    ):
        sbcount += 1

        args = {"S": s[1], "M": m[1], "T": t[1], "Z": z[1], "C": c[1]}

        # adjust arguments if dimensions are not present
        if not has_t:
            args.pop("T")
        if not has_z:
            args.pop("Z")
        if not has_s:
            args.pop("S")
        if not has_m:
            args.pop("M")
        if not has_c:
            args.pop("C")

        # read information from subblock and bounding box
        if has_m:
            # get x, y, width and height for a specific tile
            bbox = aicsczi.get_mosaic_tile_bounding_box(**args)
            sb = aicsczi.read_subblock_metadata(unified_xml=True, B=0, S=s[1], M=m[1], T=t[1], Z=z[1], C=c[1])
        elif not has_m:
            bbox = aicsczi.get_tile_bounding_box(**args)
            sb = aicsczi.read_subblock_metadata(unified_xml=True, B=0, S=s[1], T=t[1], Z=z[1], C=c[1])

        # get information from subblock
        timestamp, xpos, ypos, zpos = _getsbinfo(sb)

        # assemble data for a single plane
        plane = [
            {
                "Subblock": sbcount,
                "S": s[1],
                "M": m[1],
                "T": t[1],
                "C": c[1],
                "Z": z[1],
                "X[micron]": xpos,
                "Y[micron]": ypos,
                "Z[micron]": zpos,
                "Time[s]": timestamp,
                "xstart": bbox.x,
                "ystart": bbox.y,
                "width": bbox.w,
                "height": bbox.h,
            }
        ]

        # concatenate the plane data to the dataframe
        df_czi = pd.concat(
            [df_czi if not df_czi.empty else None, pd.DataFrame(plane)],
            ignore_index=True,
        )

    # normalize time stamps
    if norm_time:
        df_czi = norm_columns(df_czi, colname="Time[s]", mode="min")

    if save_table:
        try:
            planetable_savepath = save_planetable(df_czi, czifile, separator=table_separator, index=table_index)
            logger.info(f"Planetable saved successfully at: {planetable_savepath}")
            return df_czi, planetable_savepath
        except Exception as e:
            logger.error(f"Failed to save planetable: {e}")
            return df_czi, None

    return df_czi


def _getsbinfo(subblock: Any) -> Tuple[float, float, float, float]:
    """
    Extracts metadata information from a given subblock element.
    This function retrieves the acquisition timestamp, stage X position,
    stage Y position, and focus position from the provided subblock.
    If any of the elements are missing, their corresponding values are
    set to 0.0.
    Args:
        subblock (Any): The subblock element containing metadata information.
    Returns:
        Tuple[float, float, float, float]: A tuple containing:
            - timestamp (float): The acquisition time in seconds since the epoch.
            - xpos (float): The stage X position.
            - ypos (float): The stage Y position.
            - zpos (float): The focus position.
    """
    try:
        time = subblock.findall(".//AcquisitionTime")[0].text
        timestamp = dt.parse(time).timestamp()
    except IndexError as e:
        timestamp = 0.0

    try:
        xpos = np.double(subblock.findall(".//StageXPosition")[0].text)
    except IndexError as e:
        xpos = 0.0

    try:
        ypos = np.double(subblock.findall(".//StageYPosition")[0].text)
    except IndexError as e:
        ypos = 0.0

    try:
        zpos = np.double(subblock.findall(".//FocusPosition")[0].text)
    except IndexError as e:
        zpos = 0.0

    return timestamp, xpos, ypos, zpos


def norm_columns(df: pd.DataFrame, colname: str = "Time [s]", mode: str = "min") -> pd.DataFrame:
    """Normalize a specific column inside a Pandas dataframe
    Args:
        df: DataFrame
        colname: Name of the column to be normalized, defaults to 'Time [s]'
        mode: Mode of Normalization, defaults to 'min'

    Returns:
        Dataframe with normalized columns
    """

    # normalize columns according to min or max value
    if mode == "min":
        min_value = df[colname].min()
        df[colname] = np.round((df[colname] - min_value), 3)

    if mode == "max":
        max_value = df[colname].max()
        df[colname] = np.round((df[colname] - max_value), 3)

    return df


def filter_planetable(planetable: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Filters the input planetable DataFrame based on specified dimension entries.

    The function uses optional keyword arguments to filter the planetable on
    various dimensions such as scene index, time index, z-plane index, and channel index.
    If a dimension is not specified in the keyword arguments, it will not be used for filtering.
    Only valid arguments are accepted, which are 's', 't', 'z', and 'c'. An error is raised for invalid arguments.

    Args:
        planetable (pd.DataFrame): The DataFrame to be filtered.
            It should contain columns "S", "M", "T", "Z" and "C".

        **kwargs: Optional keyword arguments specifying the indices to filter on.
            These can include:
                'scene': The scene index.
                'time': The time index.
                'zplane': The z-plane index.
                'channel': The channel index.

    Returns:
        pd.DataFrame: The filtered planetable.

    Raises:
        KeyError: If a specified index is not a column in the DataFrame.
        ValueError: If an invalid keyword argument is passed.

    Examples:
        >>> planetable = pd.DataFrame({'scene': [0, 1, 1], 'time': [0, 1, 0], 'zplane': [0, 1, 1], 'channel': [0, 1, 1]})
        >>> filter_planetable(planetable, scene=1, time=0)
        scene  time  zplane  channel
        2      1     0       1      1
    """

    valid_args = ["scene", "time", "zplane", "channel"]

    # check for invalid arguments
    for k, v in kwargs.items():

        if k not in valid_args:
            raise ValueError(f"Invalid keyword argument: {k}")

    if "scene" in kwargs:
        # filter planetable for specific scene
        planetable = planetable[planetable["S"] == kwargs["scene"]]

    if "time" in kwargs:
        # filter planetable for specific timepoint
        planetable = planetable[planetable["T"] == kwargs["time"]]

    if "zplane" in kwargs:
        # filter resulting planetable pt for a specific z-plane
        planetable = planetable[planetable["Z"] == kwargs["zplane"]]

    if "channel" in kwargs:
        # filter planetable for specific channel
        planetable = planetable[planetable["C"] == kwargs["channel"]]

    # return filtered planetable
    return planetable


def save_planetable(df: pd.DataFrame, filepath: str, separator: str = ",", index: bool = True) -> str:
    """Saves a pandas dataframe as a CSV file.

    Args:
        df (pd.DataFrame): The dataframe to be saved as CSV.
        filepath (str): The filepath of the CSV file to be written.
        separator (str, optional): The separator character for the CSV file. Defaults to ','.
        index (bool, optional): Whether to include the index in the CSV file. Defaults to True.

    Returns:
        str: The filepath of the CSV file that was written.
    """
    # Generate the filepath for the planetable CSV.
    csvfile = os.path.splitext(filepath)[0] + "_planetable.csv"

    # Write the dataframe to the planetable CSV file.
    df.to_csv(csvfile, sep=separator, index=index)

    return csvfile


def expand5d(array: np.ndarray) -> np.ndarray:
    """Expands a multi-dimensional numpy array to 5 dimensions.

    Args:
        array (np.ndarray): The numpy array to be extended to 5 dimensions.

    Returns:
        np.ndarray: The 5-dimensional numpy array.
    """
    # Expand the input array along the third-to-last dimension.
    array = np.expand_dims(array, axis=-3)
    # Expand the result along the fourth-to-last dimension.
    array = np.expand_dims(array, axis=-4)
    # Expand the result along the fifth-to-last dimension.
    array5d = np.expand_dims(array, axis=-5)

    return array5d


# def clean_dict(d: Dict) -> Dict:
#     """
#     Recursively cleans a dictionary by removing keys with values that are None, empty lists, or empty dictionaries.
#     Args:
#         d (Dict): The dictionary to be cleaned.
#     Returns:
#         Dict: A new dictionary with the same structure as the input, but without keys that have None, empty lists, or empty dictionaries as values.
#     """

#     def _clean_dict(d: Dict) -> Dict:
#         # Initialize an empty dictionary to store cleaned key-value pairs
#         cleaned = {}

#         # Iterate over each key-value pair in the dictionary
#         for k, v in d.items():

#             # Check if the value is a dictionary
#             if isinstance(v, dict):

#                 # Recursively clean the nested dictionary
#                 nested = _clean_dict(v)

#                 # If the nested dictionary is not empty
#                 if nested:
#                     # Add the cleaned nested dictionary to the cleaned dictionary
#                     cleaned[k] = nested

#             # Check if the value is an array and ensure it is not empty
#             elif isinstance(v, (np.ndarray, da.Array)):
#                 if v.size > 0:
#                     cleaned[k] = v

#             # Check if the value is not None, an empty list, or an empty dictionary
#             elif v is not None and not isinstance(v, (np.ndarray, da.Array)) and v != [] and v != {}:
#                 cleaned[k] = v

#         return cleaned

#     return _clean_dict(d)  # Call the inner function and return its result


def clean_dict(d: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Recursively cleans a dictionary by removing keys with values that are None, empty lists, empty dictionaries,
    or empty NumPy arrays.

    Args:
        d (Dict): The dictionary to be cleaned.

    Returns:
        Dict: A new dictionary with the same structure as the input, but without keys that have None, empty lists,
        empty dictionaries, or empty NumPy arrays as values.
    """
    if not isinstance(d, dict):
        raise ValueError("Input must be a dictionary.")

    cleaned = {}
    for key, value in d.items():
        # Handle None values
        if value is None:
            continue

        # Handle empty lists
        if isinstance(value, list) and len(value) == 0:
            continue

        # Handle empty dictionaries
        if isinstance(value, dict) and len(value) == 0:
            continue

        # Handle empty NumPy arrays
        if isinstance(value, np.ndarray) and value.size == 0:
            continue

        # Recursively clean nested dictionaries
        if isinstance(value, dict):
            nested_cleaned = clean_dict(value)
            if nested_cleaned:  # Only include if the cleaned nested dictionary is not empty
                cleaned[key] = nested_cleaned
        else:
            cleaned[key] = value

    return cleaned


def download_zip(source_link: str) -> str:

    if not validators.url(source_link):
        logger.warning("Not a valid link.")
        return ""

    import io
    import zipfile

    response = requests.get(source_link, stream=True)
    compressed_data = io.BytesIO(response.content)

    with zipfile.ZipFile(compressed_data, "r") as zip_accessor:
        zip_accessor.extractall("./")

    return compressed_data[:-4]


def check_zoom(zoom: Annotated[float, ValueRange(0.01, 1.0)] = 1.0) -> float:

    # check zoom factor
    if zoom > 1.0:
        logger.warning(f"Zoom factor f{zoom} is not in valid range [0.01 - 1.0]. Using 1.0 instead.")
        zoom = 1.0
    if zoom < 0.01:
        logger.warning(f"Zoom factor f{zoom} is not in valid range [0.01 - 1.0]. Using 0.01 instead.")
        zoom = 0.01

    return zoom


def measure_memory_usage(target_function):
    """
    A decorator that measures and logs the memory usage of the decorated function.
    This decorator uses the `tracemalloc` module to track memory allocations and logs
    the top memory-consuming lines after the function execution.
    Args:
        target_function (function): The function to be decorated.
    Returns:
        function: The wrapped function with memory usage measurement.
    Example:
        @measure_memory_usage
        def my_function():
            # Function implementation
            pass
    """

    def wrapper(*args, **kwargs):
        tracemalloc.start()

        # Call the original function
        result = target_function(*args, **kwargs)

        snapshot = tracemalloc.take_snapshot()

        top_stats = snapshot.statistics("lineno")
        # # Print the top memory-consuming lines
        logger.info(f"Memory usage of {target_function.__name__}:")
        for stat in top_stats[:3]:
            logger.info(stat)

        # top_stats = snapshot.statistics("traceback")
        ## pick the biggest memory block
        # stat = top_stats[0]
        # print("%s memory blocks: %.1f KiB" % (stat.count, stat.size / 1024))
        # for line in stat.traceback.format():
        #    logger.info(line)

        # Return the result
        return result

    return wrapper


def measure_execution_time(func):
    """
    Decorator that measures the execution time of a function and logs it.
    Args:
        func (callable): The function to be decorated.
    Returns:
        callable: The wrapped function with execution time measurement.
    Example:
        @measure_execution_time
        def my_function():
            # Function implementation
            pass
    The execution time will be logged using the logger with an info level.
    """

    def timed_execution(*args, **kwargs):
        start_timestamp = time.time()
        result = func(*args, **kwargs)
        end_timestamp = time.time()
        execution_duration = end_timestamp - start_timestamp
        logger.info(f"Function: {func.__name__} --> {execution_duration:.2f} [s]")
        return result

    return timed_execution



# -*- coding: utf-8 -*-

#################################################################
# File        : read_tools.py
# Author      : sebi06
#
# Disclaimer: This code is purely experimental. Feel free to
# use it at your own risk.
#
#################################################################

from typing import Dict, Tuple, Optional, Union, List
from pylibCZIrw import czi as pyczi
from aicspylibczi import CziFile
from czitools.metadata_tools import czi_metadata as czimd
from czitools.utils import misc
import numpy as np
from pathlib import Path
import dask.array as da
import dask.delayed
import os
from tqdm import tqdm
from tqdm.contrib.itertools import product
import tempfile
import shutil
from czitools.utils import logging_tools
from czitools.metadata_tools.helper import AttachmentType
import xarray as xr

# from memory_profiler import profile

logger = logging_tools.set_logging()


# code for which memory has to be monitored
# instantiating the decorator
# @profile
def read_6darray(
    filepath: Union[str, os.PathLike[str]],
    use_dask: Optional[bool] = False,
    chunk_zyx: Optional[bool] = False,
    planes: Optional[Dict[str, Tuple[int, int]]] = None,
    zoom: Optional[float] = 1.0,
    use_xarray: Optional[bool] = True,
) -> Tuple[Optional[Union[np.ndarray, da.Array, xr.DataArray]], czimd.CziMetadata]:
    """Read a CZI image file as 6D dask array.
    Important: Currently supported are only scenes with equal size and CZIs with consistent pixel types.
    The output array has always the dimension order: STCZYX(A)

    Args:
        filepath (str | Path): filepath for the CZI image
        use_dask (Optional[bool], optional): Option to store image data as dask array with delayed reading. Defaults to False.
        chunk_zyx (Optional[bool], optional): Option to chunk dask array along z-stacks. Defaults to False.
        planes (Optional[Dict[str, Tuple[int, int]]], optional): Allowed keys S, T, Z, C and their start and end values must be >=0 (zero-based).
                                 planes = {"Z":(0, 2)} will return 3 z-plane with indices (0, 1, 2).
                                 Respectively {"Z":(5, 5)} will return a single z-plane with index 5. Defaults to None.
        zoom (Optional[float], optional): Downscale images using a factor [0.01 - 1.0]. Defaults to 1.0.
        use_xarray (Optional[bool], optional): Option to use xarray for the output array. Defaults to True.

    Returns:
        Tuple[array6d, mdata]: output as 6D numpy, dask or xarray and metadata_tools
    """

    if isinstance(filepath, Path):
        # convert to string
        filepath = str(filepath)

    # check zoom factor for valid range
    zoom = misc.check_zoom(zoom=zoom)

    # define used dimensions
    dims = ("S", "T", "C", "Z", "Y", "X", "A")

    # get the complete metadata_tools at once as one big class
    mdata = czimd.CziMetadata(filepath)

    if not mdata.consistent_pixeltypes:
        logger.info("Detected PixelTypes ar not consistent. Cannot create array6d")
        return None, mdata

    # update scaling
    mdata.scale.X_sf = np.round(mdata.scale.X * (1 / zoom), 3)
    mdata.scale.Y_sf = np.round(mdata.scale.Y * (1 / zoom), 3)
    mdata.scale.ratio["zx_sf"] = np.round(mdata.scale.Z / mdata.scale.X_sf, 3)

    # check planes
    if planes:
        for k in ["S", "T", "C", "Z"]:
            if k in planes.keys() and k in mdata.bbox.total_bounding_box.keys():
                if mdata.bbox.total_bounding_box[k][1] - 1 < planes[k][1]:
                    logger.info(
                        f"Planes indices (zero-based) for {planes[k]} are invalid. BBox for {[k]}: {mdata.bbox.total_bounding_box[k]}"
                    )
                    return None, mdata

    if mdata.consistent_pixeltypes:
        # use pixel type from first channel
        use_pixeltype = mdata.npdtype_list[0]

    if not mdata.scene_shape_is_consistent and "S" not in (planes or {}):
        logger.info("Scenes have inconsistent shape. Cannot read 6D array")
        return None, mdata

    # open the CZI document to read the
    with pyczi.open_czi(filepath, mdata.pyczi_readertype) as czidoc:

        # check if dimensions are None (because they do not exist for that image)
        size_c = misc.check_dimsize(mdata.image.SizeC, set2value=1)
        size_z = misc.check_dimsize(mdata.image.SizeZ, set2value=1)
        size_t = misc.check_dimsize(mdata.image.SizeT, set2value=1)
        size_s = misc.check_dimsize(mdata.image.SizeS, set2value=1)

        s_start = 0
        s_end = size_s
        t_start = 0
        t_end = size_t
        c_start = 0
        c_end = size_c
        z_start = 0
        z_end = size_z

        # check for additional arguments to create substacks
        if planes and mdata.image.SizeS is not None and "S" in planes.keys():
            size_s = planes["S"][1] - planes["S"][0] + 1
            mdata.image.SizeS = size_s
            s_start = planes["S"][0]
            s_end = planes["S"][1] + 1

        if planes and mdata.image.SizeT is not None and "T" in planes.keys():
            size_t = planes["T"][1] - planes["T"][0] + 1
            mdata.image.SizeT = size_t
            t_start = planes["T"][0]
            t_end = planes["T"][1] + 1

        if planes and mdata.image.SizeZ is not None and "Z" in planes.keys():
            size_z = planes["Z"][1] - planes["Z"][0] + 1
            mdata.image.SizeZ = size_z
            z_start = planes["Z"][0]
            z_end = planes["Z"][1] + 1

        if planes and mdata.image.SizeC is not None and "C" in planes.keys():
            size_c = planes["C"][1] - planes["C"][0] + 1
            mdata.image.SizeC = size_c
            c_start = planes["C"][0]
            c_end = planes["C"][1] + 1

        # check if ADim can be removed because image is grayscale
        # Check if any value in the dictionary is True
        contains_rgb = any(mdata.isRGB.values())
        remove_adim = False if contains_rgb else True

        if mdata.is_url:
            logger.info("Reading pixel data via network from link location.")

        planecount = 0
        shape = ()

        # read array for the scene 2Dplane-by-2Dplane
        for s, t, c, z in product(
            enumerate(range(s_start, s_end)),
            enumerate(range(t_start, t_end)),
            enumerate(range(c_start, c_end)),
            enumerate(range(z_start, z_end)),
            desc="Reading 2D planes",
            unit=" 2Dplanes",
        ):
            planecount += 1

            # read a 2D image plane from the CZI
            if mdata.image.SizeS is None:
                image2d = czidoc.read(plane={"T": t[1], "Z": z[1], "C": c[1]}, zoom=zoom)
            else:
                image2d = czidoc.read(plane={"T": t[1], "Z": z[1], "C": c[1]}, scene=s[1], zoom=zoom)

            if planecount == 1:
                # allocate array based on the expected size incl. down scaling
                shape = (
                    size_s,
                    size_t,
                    size_c,
                    size_z,
                    image2d.shape[0],
                    image2d.shape[1],
                    3 if contains_rgb else 1,
                )

                # in case of numpy array
                if not use_dask:
                    array6d = np.empty(shape, dtype=use_pixeltype)

                # in case of normal dask array
                elif use_dask:
                    array6d = da.empty(shape, dtype=use_pixeltype, chunks=shape)

            # insert 2D image plane into the array6d
            array6d[s[0], t[0], c[0], z[0], ...] = image2d

        # remove the A dimension
        if remove_adim:
            array6d = np.squeeze(array6d, axis=-1)
            dims = ("S", "T", "C", "Z", "Y", "X")

            if use_dask and chunk_zyx:
                # re-chunk array based on shape
                array6d = array6d.rechunk(chunks=(1, 1, 1, size_z, image2d.shape[0], image2d.shape[1]))

        if not remove_adim:
            if use_dask and chunk_zyx:
                # re-chunk array based on shape
                array6d = array6d.rechunk(chunks=(1, 1, 1, size_z, image2d.shape[0], image2d.shape[1], 3))

    # update metadata_tools
    mdata.array6d_size = array6d.shape

    if use_xarray:

        # create xarray from array6d
        coords = {}
        for index, dim in enumerate(dims):
            # Create a range for each dimension
            coords[dim] = range(array6d.shape[index])

        # Create the xarray.DataArray
        array6d = xr.DataArray(array6d, dims=dims, coords=coords)

        # Set attributes for the DataArray
        array6d.attrs = {
            "description": "6D image data from CZI file",
            "source": mdata.filepath,
            "axes": "".join(dims),
            # "metadata": mdata,  # Include metadata if it's a dictionary or serializable
        }

    return array6d, mdata


# code for which memory has to be monitored
# instantiating the decorator
# @profile
def read_6darray_lazy(
    filepath: Union[str, os.PathLike[str]],
    chunk_zyx=False,
    planes: Optional[Dict[str, Tuple[int, int]]] = None,
    zoom: Optional[float] = 1.0,
) -> Tuple[Optional[Union[np.ndarray, da.Array]], czimd.CziMetadata]:
    """Read a CZI image file as 6D dask array with delayed plane reading.
    Important: Currently supported are only scenes with equal size and CZIs with consistent pixel types.
    The output array has always the dimension order: STCZYX(A)

    Args:
        filepath (str | Path): filepath for the CZI image
        chunk_zyx (bool, optional): Option to chunk dask array along z-stacks. Defaults to False.
        zoom (float, optional): Downscale images using a factor [0.01 - 1.0]. Defaults to 1.0
        planes (dict, optional): Allowed keys S, T, Z, C and their start and end values must be >=0 (zero-based).
                                 planes = {"Z":(0, 2)} will return 3 z-plane with indices (0, 1, 2).
                                 Respectively {"Z":(5, 5)} will return a single z-plane with index 5.

    Returns:
        Tuple[array6d, mdata]: output as 6D dask array and metadata_tools
    """

    if isinstance(filepath, Path):
        # convert to string
        filepath = str(filepath)

    # get the complete metadata_tools at once as one big class
    mdata = czimd.CziMetadata(filepath)

    if not mdata.consistent_pixeltypes:
        logger.info("Detected PixelTypes are not consistent. Cannot create array6d")
        return None, mdata

    # check zoom factor for valid range
    zoom = misc.check_zoom(zoom=zoom)

    # update scaling
    mdata.scale.X_sf = np.round(mdata.scale.X * (1 / zoom), 3)
    mdata.scale.Y_sf = np.round(mdata.scale.Y * (1 / zoom), 3)
    mdata.scale.ratio["zx_sf"] = np.round(mdata.scale.Z / mdata.scale.X_sf, 3)

    # check planes
    if planes:
        for k in ["S", "T", "C", "Z"]:
            if k in planes.keys() and k in mdata.bbox.total_bounding_box.keys():
                if mdata.bbox.total_bounding_box[k][1] - 1 < planes[k][1]:
                    logger.info(
                        f"Planes indices (zero-based) for {planes[k]} are invalid. BBox for {[k]}: {mdata.bbox.total_bounding_box[k]}"
                    )
                    return None, mdata

    if not mdata.scene_shape_is_consistent and "S" not in (planes or {}):
        logger.info("Scenes have inconsistent shape. Cannot read 6D array")
        return None, mdata

    # open the CZI document to read
    with pyczi.open_czi(filepath, mdata.pyczi_readertype) as czidoc:
        if mdata.image.SizeS is not None:
            # use the size of the 1st scenes_bounding_rectangle
            size_x = czidoc.scenes_bounding_rectangle[0].w
            size_y = czidoc.scenes_bounding_rectangle[0].h
        else:
            # use the size of the total_bounding_rectangle
            size_x = czidoc.total_bounding_rectangle.w
            size_y = czidoc.total_bounding_rectangle.h

        # check if dimensions are None (because they do not exist for that image)
        size_c = misc.check_dimsize(mdata.image.SizeC, set2value=1)
        size_z = misc.check_dimsize(mdata.image.SizeZ, set2value=1)
        size_t = misc.check_dimsize(mdata.image.SizeT, set2value=1)
        size_s = misc.check_dimsize(mdata.image.SizeS, set2value=1)

        s_start, s_end = 0, size_s
        t_start, t_end = 0, size_t
        c_start, c_end = 0, size_c
        z_start, z_end = 0, size_z

        # check for additional arguments to create substacks
        if planes and mdata.image.SizeS is not None and "S" in planes.keys():
            size_s = planes["S"][1] - planes["S"][0] + 1
            mdata.image.SizeS = size_s
            s_start = planes["S"][0]
            s_end = planes["S"][1] + 1

        if planes and mdata.image.SizeT is not None and "T" in planes.keys():
            size_t = planes["T"][1] - planes["T"][0] + 1
            mdata.image.SizeT = size_t
            t_start = planes["T"][0]
            t_end = planes["T"][1] + 1

        if planes and mdata.image.SizeZ is not None and "Z" in planes.keys():
            size_z = planes["Z"][1] - planes["Z"][0] + 1
            mdata.image.SizeZ = size_z
            z_start = planes["Z"][0]
            z_end = planes["Z"][1] + 1

        if planes and mdata.image.SizeC is not None and "C" in planes.keys():
            size_c = planes["C"][1] - planes["C"][0] + 1
            mdata.image.SizeC = size_c
            c_start = planes["C"][0]
            c_end = planes["C"][1] + 1

        # check if the A Dimension can be removed because image is grayscale
        contains_rgb = any(mdata.isRGB.values())
        remove_adim = not contains_rgb

        shape2d = (size_y, size_x, 3) if contains_rgb else (size_y, size_x)

        # initialize empty list to hold the dask arrays
        img = []

        with tqdm(
            total=size_s * size_t * size_c * size_z,
            desc="Reading 2D planes",
            unit=" 2Dplanes",
        ) as pbar:
            for s in range(s_start, s_end):
                time_stack = []

                for t in range(t_start, t_end):
                    ch_stack = []

                    for c in range(c_start, c_end):
                        z_stack = []

                        for z in range(z_start, z_end):
                            z_slice = da.from_delayed(
                                read_2dplane(
                                    czidoc,
                                    s=s,
                                    t=t,
                                    c=c,
                                    z=z,
                                    has_scenes=mdata.has_scenes,
                                    zoom=zoom,
                                    remove_adim=remove_adim,
                                ),
                                shape=shape2d,
                                dtype=mdata.npdtype[0],
                            )
                            pbar.update()

                            # append to the z-stack
                            z_stack.append(z_slice)

                        # stack the array and create new dask array
                        z_stack = da.stack(z_stack, axis=0)
                        ch_stack.append(z_stack)

                    time_stack.append(ch_stack)

                img.append(time_stack)

            # create final STCZYX dask array
            array6d = da.stack(img, axis=0)

        # remove the A dimension
        if remove_adim:
            if chunk_zyx:
                array6d = array6d.rechunk(chunks=(1, 1, 1, size_z, size_y, size_x))

        if not remove_adim:
            if chunk_zyx:
                array6d = array6d.rechunk(chunks=(1, 1, 1, size_z, size_y, size_x, 3))

    return array6d, mdata


@dask.delayed
def read_2dplane(
    czidoc: pyczi.CziReader,
    s: int = 0,
    t: int = 0,
    c: int = 0,
    z: int = 0,
    has_scenes: bool = True,
    zoom: float = 1.0,
    remove_adim: bool = True,
):
    """Dask delayed function to read a 2d plane from a CZI image, which has the shape
       (Y, X, 1) or (Y, X, 3).
       If the image is a grayscale image the last array dimension will be removed, when
       the option is ste to True.

    Args:
        czidoc (pyczi.CziReader): CziReader objects
        s (int, optional): Scene index. Defaults to 0.
        t (int, optional): Time index. Defaults to 0.
        c (int, optional): Channel index. Defaults to 0.
        z (int, optional): Z-Plane index. Defaults to 0.
        has_scenes (bool, optional): Defines if the CZI actually contains scenes. Defaults to True.
        zoom (float, optional): Downscale images using a factor [0.01 - 1.0]. Defaults to 1.0
        remove_adim (bool, optional): Option to remove the last dimension of the 2D array. Defaults to True.

    Returns:
        dask.array: 6d dask.array with delayed reading for individual 2d planes
    """

    # initialize 2d array with some values
    image2d = np.zeros([10, 10], dtype=np.int16)

    if has_scenes:
        # read a 2d plane using the scene index
        image2d = czidoc.read(plane={"T": t, "C": c, "Z": z}, scene=s, zoom=zoom)
    if not has_scenes:
        # reading a 2d plane in case the CZI has no scenes
        image2d = czidoc.read(plane={"T": t, "C": c, "Z": z}, zoom=zoom)

    # remove a last "A" dimension when desired
    if remove_adim:
        return image2d[..., 0]
    if not remove_adim:
        return image2d


def read_attachments(
    czi_filepath: Union[str, os.PathLike],
    attachment_type: AttachmentType = AttachmentType.SlidePreview,
    copy: bool = True,
) -> Tuple[Optional[np.ndarray], Optional[str]]:
    """Read attachment images from a CZI image as numpy array

    Args:
        czi_filepath (str): FilePath of the CZI image file
        attachment_type (str, optional): Type of the attachment to be read. Defaults to "SlidePreview".
        copy (bool, optional): Option to copy the attachments as CZI image directly. Defaults to True.

    Raises:
        Exception: If specified attachment is not found and exception is raised.

    Returns:
        Tuple[nd,array, [Optional[str]]: Tuple containing the 2d image array and optionally the location of the copied image.
    """

    try:
        import czifile

        if attachment_type not in AttachmentType:
            # if attachment_type not in ["SlidePreview", "Label", "]:
            raise Exception(f"{attachment_type} is not supported. Valid types are: SlidePreview, Label or Prescan.")

        att = czimd.CziAttachments(czi_filepath)

        if attachment_type == AttachmentType.Label and not att.has_label:
            # if attachment_type == "Label" and not att.has_label:
            return np.array([]), None

        if attachment_type == AttachmentType.SlidePreview and not att.has_preview:
            # if attachment_type == "SlidePreview" and not att.has_preview:
            return np.array([]), None

        if attachment_type == AttachmentType.Prescan and not att.has_prescan:
            # if attachment_type == "SlidePreview" and not att.has_preview:
            return np.array([]), None

        # create CZI-object using czifile library
        with czifile.CziFile(czi_filepath) as cz:
            with tempfile.TemporaryDirectory() as tmpdirname:
                # save attachments to temporary directory
                cz.save_attachments(directory=tmpdirname)

                # iterate over attachments
                for att in cz.attachments():
                    if att.attachment_entry.name == attachment_type.name:
                        # get the full path of the attachment image
                        full_path = Path(tmpdirname) / att.attachment_entry.filename

                        if copy:
                            # create path to store the attachment image
                            att_path = str(czi_filepath)[:-4] + "_" + att.attachment_entry.name + ".czi"

                            # copy the file
                            dest = shutil.copyfile(full_path, att_path)

                        # open the CZI document to read array
                        with pyczi.open_czi(str(full_path)) as czidoc:
                            img2d = czidoc.read()

                        if copy:
                            return img2d, dest
                        if not copy:
                            return img2d

    except ImportError:  # as e:
        logger.warning("Package czifile not found. Cannot extract information about attached images.")

        return None, None


def read_tiles(filepath: Union[str, os.PathLike[str]], scene: int, tile: int, **kwargs) -> Tuple[np.ndarray, List]:
    """
    Reads a specific tile from a CZI file.
    Parameters:
    -----------
    filepath : Union[str, os.PathLike[str]]
        Path to the CZI file.
    scene : int
        The scene index to read from the CZI file.
    tile : int
        The tile index to read from the CZI file.
    **kwargs : dict
        Additional keyword arguments to specify substacks. Valid arguments are:
        - 'T': Time dimension
        - 'Z': Z-dimension (depth)
        - 'C': Channel dimension
    Returns:
    --------
    Tuple[np.ndarray, List]
        A tuple containing:
        - tile_stack (np.ndarray): The image data of the specified tile.
        - size (List): A list of tuples representing the dimensions and their sizes.
    Raises:
    -------
    ValueError
        If an invalid keyword argument is provided in **kwargs.
    Notes:
    ------
    This function uses the `aicspylibczi` library to read CZI files. If the CZI file
    is a mosaic, the M-dimension is handled accordingly.
    """

    if isinstance(filepath, Path):
        # convert to string
        filepath = str(filepath)

    valid_args = ["T", "Z", "C"]

    # check for invalid arguments to specify substacks
    for k, v in kwargs.items():
        if k not in valid_args:
            raise ValueError(f"Invalid keyword argument: {k}")

    # read CZI using aicspylibczi: : https://pypi.org/project/aicspylibczi/
    czi = CziFile(filepath)

    # show the values
    logger.info(f"Reading File: {filepath} Scene: {scene} - Tile {tile}")
    logger.info(f"Dimensions Shape: {czi.get_dims_shape()}")

    # in case the CZI is a mosaic file and has the M-dimension
    if czi.is_mosaic():
        tile_stack, size = czi.read_image(S=scene, M=tile, **kwargs)

        # remove the M-Dimension from the array and size
        tile_stack = np.squeeze(tile_stack, axis=czi.dims.find("M"))
        size.remove(("M", 1))

    # in case the CZI is not a mosaic file and has no M-dimension
    elif not czi.is_mosaic():
        logger.warning("CZI file is not a mosaic. No M-Dimension found.")
        tile_stack, size = czi.read_image(S=scene, **kwargs)

    return tile_stack, size



from __future__ import annotations
from typing import Union
import os
from pylibCZIrw import czi as pyczi
from box import Box
import validators


def get_czimd_box(filepath: Union[str, os.PathLike[str]]) -> Box:
    """
    get_czimd_box: Get CZI metadata_tools as a python-box. For details: https://pypi.org/project/python-box/

    Args:
        filepath (Union[str, os.PathLike[str]]): Filepath of the CZI file

    Returns:
        Box: CZI metadata_tools as a Box object
    """

    readertype = pyczi.ReaderFileInputTypes.Standard

    if validators.url(str(filepath)):
        readertype = pyczi.ReaderFileInputTypes.Curl

    # get metadata_tools dictionary using pylibCZIrw
    with pyczi.open_czi(str(filepath), readertype) as czi_document:
        metadata_dict = czi_document.metadata
        # total_bounding_box_no_pyramid = czi_document.total_bounding_box_no_pyramid
        scenes_bounding_rectangle = czi_document.scenes_bounding_rectangle

    czimd_box = Box(
        metadata_dict,
        conversion_box=True,
        default_box=True,
        default_box_attr=None,
        default_box_create_on_get=True,
        # default_box_no_key_error=True
    )

    # add the filepath
    czimd_box.filepath = filepath
    czimd_box.is_url = validators.url(str(filepath))
    czimd_box.czi_open_arg = readertype

    # set the defaults to False
    czimd_box.has_customattr = False
    czimd_box.has_experiment = False
    czimd_box.has_disp = False
    czimd_box.has_hardware = False
    czimd_box.has_scale = False
    czimd_box.has_instrument = False
    czimd_box.has_microscopes = False
    czimd_box.has_detectors = False
    czimd_box.has_objectives = False
    czimd_box.has_tubelenses = False
    czimd_box.has_disp = False
    czimd_box.has_channels = False
    czimd_box.has_info = False
    czimd_box.has_app = False
    czimd_box.has_doc = False
    czimd_box.has_image = False
    czimd_box.has_scenes = False
    czimd_box.has_T = False
    czimd_box.has_Z = False
    czimd_box.has_dims = False
    czimd_box.has_layers = False

    if "Experiment" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_experiment = True

    if "HardwareSetting" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_hardware = True

    if "CustomAttributes" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_customattr = True

    if "Information" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_info = True

        if "Application" in czimd_box.ImageDocument.Metadata.Information:
            czimd_box.has_app = True

        if "Document" in czimd_box.ImageDocument.Metadata.Information:
            czimd_box.has_doc = True

        if "Image" in czimd_box.ImageDocument.Metadata.Information:
            czimd_box.has_image = True

            if "Dimensions" in czimd_box.ImageDocument.Metadata.Information.Image:
                czimd_box.has_dims = True

                if (
                    "Channels"
                    in czimd_box.ImageDocument.Metadata.Information.Image.Dimensions
                ):
                    # if "C" in total_bounding_box_no_pyramid.keys():
                    czimd_box.has_channels = True

                if "T" in czimd_box.ImageDocument.Metadata.Information.Image.Dimensions:
                    # if "T" in total_bounding_box_no_pyramid.keys():
                    czimd_box.has_T = True

                if "Z" in czimd_box.ImageDocument.Metadata.Information.Image.Dimensions:
                    # if "Z" in total_bounding_box_no_pyramid.keys():
                    czimd_box.has_Z = True

                # if "S" in czimd_box.ImageDocument.Metadata.Information.Image.Dimensions:
                if len(scenes_bounding_rectangle) > 0:
                    czimd_box.has_scenes = True

        if "Instrument" in czimd_box.ImageDocument.Metadata.Information:
            czimd_box.has_instrument = True

            if "Detectors" in czimd_box.ImageDocument.Metadata.Information.Instrument:
                czimd_box.has_detectors = True

            if "Microscopes" in czimd_box.ImageDocument.Metadata.Information.Instrument:
                czimd_box.has_microscopes = True

            if "Objectives" in czimd_box.ImageDocument.Metadata.Information.Instrument:
                czimd_box.has_objectives = True

            if "TubeLenses" in czimd_box.ImageDocument.Metadata.Information.Instrument:
                czimd_box.has_tubelenses = True

    if "Scaling" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_scale = True

    if "DisplaySetting" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_disp = True

    if "Layers" in czimd_box.ImageDocument.Metadata:
        czimd_box.has_layers = True

    return czimd_box



# -*- coding: utf-8 -*-

#################################################################
# File        : vis_tools.py
# Author      : sebi06
#
# Disclaimer: This code is purely experimental. Feel free to
# use it at your own risk.
#
#################################################################

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import plotly.graph_objects as go
from czitools.utils import logging_tools

logger = logging_tools.set_logging()


def scatterplot_mpl(
    planetable,
    s=0,
    t=0,
    z=0,
    c=0,
    msz2d=35,
    normz=True,
    fig1savename="zsurface2d.png",
    fig2savename="zsurface3d.png",
    msz3d=20,
):
    """
    Generates 2D and 3D scatter plots of XYZ positions from a given table and saves them as PNG files.
    Parameters:
    -----------
    planetable : pandas.DataFrame
        DataFrame containing the XYZ positions with columns 'X[micron]', 'Y[micron]', and 'Z[micron]'.
    s : int, optional
        Unused parameter, default is 0.
    t : int, optional
        Time point, default is 0.
    z : int, optional
        Z position, default is 0.
    c : int, optional
        Channel, default is 0.
    msz2d : int, optional
        Marker size for the 2D scatter plot, default is 35.
    normz : bool, optional
        If True, normalize Z data by subtracting the minimum value, default is True.
    fig1savename : str, optional
        Filename for saving the 2D scatter plot, default is "zsurface2d.png".
    fig2savename : str, optional
        Filename for saving the 3D scatter plot, default is "zsurface3d.png".
    msz3d : int, optional
        Marker size for the 3D scatter plot, default is 20.
    Returns:
    --------
    fig1 : matplotlib.figure.Figure
        The 2D scatter plot figure.
    fig2 : matplotlib.figure.Figure
        The 3D scatter plot figure.
    Raises:
    -------
    KeyError
        If the required columns 'X[micron]', 'Y[micron]', or 'Z[micron]' are not found in the DataFrame.
    Notes:
    ------
    - The function attempts to find a "good" aspect ratio for the figures based on the data.
    - The Y-axis is inverted to have the origin (0,0) at the top-left.
    - The color of the scatter points represents the Z positions.
    - The function logs the filenames of the saved figures.
    """

    # extract XYZ positions
    try:
        xpos = planetable["X[micron]"]
        ypos = planetable["Y[micron]"]
        zpos = planetable["Z[micron]"]
    except KeyError as e:
        xpos = planetable["X [micron]"]
        ypos = planetable["Y [micron]"]
        zpos = planetable["Z [micron]"]

    # normalize z-data by substracting the minimum value
    if normz:
        zpos = zpos - zpos.min()

    # create a name for the figure
    figtitle = "XYZ-Positions: T=" + str(t) + " Z=" + str(z) + " CH=" + str(c)

    # try to find a "good" aspect ratio for the figures
    dx = xpos.max() - xpos.min()
    dy = ypos.max() - ypos.min()
    fsy = 8
    fsx = int(np.ceil(fsy * dx / dy))

    # create figure
    fig1, ax1 = plt.subplots(1, 1, figsize=(fsx + 1, fsy))

    # invert the Y-axis --> O,O = Top-Left
    ax1.invert_yaxis()

    # configure the axis
    ax1.set_title(figtitle)
    ax1.set_xlabel("Stage X-Axis [micron]", fontsize=12, fontweight="normal")
    ax1.set_ylabel("Stage Y-Axis [micron]", fontsize=12, fontweight="normal")
    ax1.grid(True)
    ax1.set_aspect("equal", "box")

    # plot data and label the colorbar
    sc1 = ax1.scatter(
        xpos,
        ypos,
        marker="s",
        c=zpos,
        s=msz2d,
        facecolor=cm.coolwarm,
        edgecolor="black",
    )

    # add the colorbar on the right-hand side
    cb1 = plt.colorbar(sc1, fraction=0.046, shrink=0.8, pad=0.04)

    # add a label
    if normz:
        cb1.set_label(
            "Z-Offset [micron]", labelpad=20, fontsize=12, fontweight="normal"
        )
    if not normz:
        cb1.set_label(
            "Z-Position [micron]", labelpad=20, fontsize=12, fontweight="normal"
        )

    # save figure as PNG
    fig1.savefig(fig1savename, dpi=100)
    logger.info(f"Saved: {fig1savename}")

    # 3D plot of surface
    fig2 = plt.figure(figsize=(fsx + 1, fsy))
    ax2 = fig2.add_subplot(111, projection="3d")

    # invert the Y-axis --> O,O = Top-Left
    ax2.invert_yaxis()

    # define the labels
    ax2.set_xlabel("Stage X-Axis [micron]", fontsize=12, fontweight="normal")
    ax2.set_ylabel("Stage Y-Axis [micron]", fontsize=12, fontweight="normal")
    ax2.set_title(figtitle)

    # plot data and label the colorbar
    sc2 = ax2.scatter(
        xpos,
        ypos,
        zpos,
        marker=".",
        s=msz3d,
        c=zpos,
        facecolor=cm.coolwarm,
        depthshade=False,
    )

    # add colorbar to the 3d plot
    cb2 = plt.colorbar(sc2, shrink=0.8)
    # add a label
    if normz:
        cb2.set_label(
            "Z-Offset [micron]", labelpad=20, fontsize=12, fontweight="normal"
        )
    if not normz:
        cb2.set_label(
            "Z-Position [micron]", labelpad=20, fontsize=12, fontweight="normal"
        )

    # save figure as PNG
    fig2.savefig(fig2savename, dpi=100)
    logger.info(f"Saved: {fig2savename}")

    return fig1, fig2


def scatterplot_plotly(
    planetable,
    s=0,
    t=0,
    z=0,
    c=0,
    msz2d=35,
    normz=True,
    fig1savename="zsurface2d.html",
    fig2savename="zsurface3d.html",
    msz3d=20,
):
    """
    Generates 2D and 3D scatter plots using Plotly and saves them as HTML files.
    Parameters:
    -----------
    planetable : pandas.DataFrame
        DataFrame containing the XYZ positions and other relevant data.
    s : int, optional
        Placeholder parameter (default is 0).
    t : int, optional
        Placeholder parameter (default is 0).
    z : int, optional
        Placeholder parameter (default is 0).
    c : int, optional
        Placeholder parameter (default is 0).
    msz2d : int, optional
        Marker size for the 2D scatter plot (default is 35).
    normz : bool, optional
        If True, normalize the Z data by subtracting the minimum value (default is True).
    fig1savename : str, optional
        Filename for saving the 2D scatter plot HTML file (default is "zsurface2d.html").
    fig2savename : str, optional
        Filename for saving the 3D scatter plot HTML file (default is "zsurface3d.html").
    msz3d : int, optional
        Marker size for the 3D scatter plot (default is 20).
    Returns:
    --------
    fig1 : plotly.graph_objs._figure.Figure
        The generated 2D scatter plot figure.
    fig2 : plotly.graph_objs._figure.Figure
        The generated 3D scatter plot figure.
    """

    # extract XYZ position for the selected channel
    try:
        xpos = planetable["X[micron]"]
        ypos = planetable["Y[micron]"]
        zpos = planetable["Z[micron]"]
    except KeyError as e:
        xpos = planetable["X [micron]"]
        ypos = planetable["Y [micron]"]
        zpos = planetable["Z [micron]"]

    # normalize z-data by substracting the minimum value
    if normz:
        zpos = zpos - zpos.min()
        scalebar_title = "Z-Offset [micron]"
    if not normz:
        scalebar_title = "Z-Position [micron]"

    # create a name for the figure
    figtitle = "XYZ-Positions: T=" + str(t) + " Z=" + str(z) + " CH=" + str(c)

    fig1 = go.Figure(
        data=go.Scatter(
            x=xpos,
            y=ypos,
            mode="markers",
            text=np.round(zpos, 1),
            marker_symbol="square",
            marker_size=msz2d,
            marker=dict(
                color=zpos,
                colorscale="Viridis",
                line_width=2,
                showscale=True,
                colorbar=dict(
                    thickness=10, title=dict(text=scalebar_title, side="right")
                ),
            ),
        )
    )

    fig1.update_xaxes(showgrid=True, zeroline=True, automargin=True)
    fig1.update_yaxes(showgrid=True, zeroline=True, automargin=True)
    fig1["layout"]["yaxis"]["autorange"] = "reversed"
    fig1.update_layout(
        title=figtitle,
        xaxis_title="StageX Position [micron]",
        yaxis_title="StageY Position [micron]",
        font=dict(size=16, color="Black"),
    )

    # save the figure
    fig1.write_html(fig1savename)
    logger.info(f"Saved: {fig1savename}")

    fig2 = go.Figure(
        data=[
            go.Scatter3d(
                x=xpos,
                y=ypos,
                z=zpos,
                mode="markers",
                marker=dict(
                    size=msz3d,
                    color=zpos,
                    colorscale="Viridis",
                    opacity=0.8,
                    colorbar=dict(
                        thickness=10, title=dict(text=scalebar_title, side="right")
                    ),
                ),
            )
        ]
    )

    fig2.update_xaxes(showgrid=True, zeroline=True, automargin=True)
    fig2.update_yaxes(showgrid=True, zeroline=True, automargin=True)
    fig2["layout"]["yaxis"]["autorange"] = "reversed"
    fig2.update_layout(
        title=figtitle,
        xaxis_title="StageX Position [micron]",
        yaxis_title="StageY Position [micron]",
        font=dict(size=16, color="Black"),
    )

    # save the figure
    fig2.write_html(fig2savename)
    logger.info(f"Saved: {fig2savename}")

    return fig1, fig2



from typing import Union, Optional, List
from dataclasses import dataclass, field
from box import Box, BoxList
import os
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box

logger = logging_tools.set_logging()


@dataclass
class CziObjectives:
    """
    CziObjectives is a class that extracts and stores objective and tubelens metadata from CZI image data.
    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): The source of the CZI image data.
        NA (List[Optional[float]]): Numerical aperture values of the objectives.
        objmag (List[Optional[float]]): Nominal magnification values of the objectives.
        Id (List[Optional[str]]): IDs of the objectives.
        name (List[Optional[str]]): Names of the objectives.
        model (List[Optional[str]]): Models of the objectives.
        immersion (List[Optional[str]]): Immersion types of the objectives.
        tubelensmag (List[Optional[float]]): Magnification values of the tubelenses.
        totalmag (List[Optional[float]]): Total magnification values calculated from objectives and tubelenses.
        verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods:
        __post_init__(): Initializes the CziObjectives instance by reading objective and tubelens information from the CZI image data.
        get_objective_info(objective: Box): Extracts and stores information from a given objective.
    """

    czisource: Union[str, os.PathLike[str], Box]
    NA: List[Optional[float]] = field(init=False, default_factory=lambda: [])
    objmag: List[Optional[float]] = field(init=False, default_factory=lambda: [])
    Id: List[Optional[str]] = field(init=False, default_factory=lambda: [])
    name: List[Optional[str]] = field(init=False, default_factory=lambda: [])
    model: List[Optional[str]] = field(init=False, default_factory=lambda: [])
    immersion: List[Optional[str]] = field(init=False, default_factory=lambda: [])
    tubelensmag: List[Optional[float]] = field(init=False, default_factory=lambda: [])
    totalmag: List[Optional[float]] = field(init=False, default_factory=lambda: [])
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading Objective Information from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        # check if objective metadata_tools actually exist
        if czi_box.has_objectives:
            try:
                # get objective data
                objective = (
                    czi_box.ImageDocument.Metadata.Information.Instrument.Objectives.Objective
                )
                if isinstance(objective, Box):
                    self.get_objective_info(objective)
                elif isinstance(objective, BoxList):
                    for obj in range(len(objective)):
                        self.get_objective_info(objective[obj])
            except AttributeError:
                objective = None

        elif not czi_box.has_objectives:
            if self.verbose:
                logger.info("No Objective Information found.")

        # check if tubelens metadata_tools exist
        if czi_box.has_tubelenses:
            # get tubelenes data
            tubelens = (
                czi_box.ImageDocument.Metadata.Information.Instrument.TubeLenses.TubeLens
            )

            if isinstance(tubelens, Box):
                if tubelens.Magnification is not None:
                    self.tubelensmag.append(float(tubelens.Magnification))
                elif tubelens.Magnification is None:
                    if self.verbose:
                        logger.warning(
                            "No tubelens magnification found. Use 1.0x instead."
                        )
                    self.tubelensmag.append(1.0)

            elif isinstance(tubelens, BoxList):
                for tl in range(len(tubelens)):
                    self.tubelensmag.append(float(tubelens[tl].Magnification))

            # some additional checks to calc the total magnification
            if self.objmag is not None and self.tubelensmag is not None:
                self.totalmag = [i * j for i in self.objmag for j in self.tubelensmag]

        elif not czi_box.has_tubelens:
            if self.verbose:
                logger.info("No Tublens Information found.")

        if self.objmag is not None and self.tubelensmag == []:
            self.totalmag = self.objmag

    def get_objective_info(self, objective: Box):
        self.name.append(objective.Name)
        self.immersion.append(objective.Immersion)

        if objective.LensNA is not None:
            self.NA.append(float(objective.LensNA))

        if objective.Id is not None:
            self.Id.append(objective.Id)

        if objective.NominalMagnification is not None:
            self.objmag.append(float(objective.NominalMagnification))

        if None in self.name and self.name.count(None) == 1:
            self.name.remove(None)
            self.name.append(objective.Manufacturer.Model)



from typing import Optional, Union, Dict
from dataclasses import dataclass, field
from box import Box
import os
from czitools.utils import logging_tools
from pylibCZIrw import czi as pyczi
import validators

logger = logging_tools.set_logging()


@dataclass
class CziBoundingBox:
    """
    A class to represent and handle bounding boxes from CZI image data.
    Attributes:
    -----------
    czisource : Union[str, os.PathLike[str], Box]
        The source of the CZI file, which can be a string, a path-like object, or a Box object.
    scenes_bounding_rect : Optional[Dict[int, pyczi.Rectangle]]
        A dictionary containing the bounding rectangles for each scene in the CZI file.
    scenes_bounding_rect_no_pyramid : Optional[Dict[int, pyczi.Rectangle]]
        A dictionary containing the bounding rectangles for each scene in the CZI file without a pyramid.
    total_rect : Optional[pyczi.Rectangle]
        The total bounding rectangle for the entire CZI file.
    total_rect_no_pyramid : Optional[pyczi.Rectangle]
        The total bounding rectangle for the entire CZI file without a pyramid.
    total_bounding_box : Optional[Dict[str, tuple]]
        A dictionary containing the total bounding box.
    total_bounding_box_no_pyramid : Optional[Dict[str, tuple]]
        A dictionary containing the total bounding box coordinates without a pyramid.
    verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods:
    --------
    __post_init__():
        Initializes the bounding box attributes by reading from the CZI file.
    """

    czisource: Union[str, os.PathLike[str], Box]
    scenes_bounding_rect: Optional[Dict[int, pyczi.Rectangle]] = field(
        init=False, default_factory=lambda: {}
    )
    scenes_bounding_rect_no_pyramid: Optional[Dict[int, pyczi.Rectangle]] = field(
        init=False, default_factory=lambda: {}
    )
    total_rect: Optional[pyczi.Rectangle] = field(init=False, default=None)
    total_rect_no_pyarmid: Optional[pyczi.Rectangle] = field(init=False, default=None)
    total_bounding_box: Optional[Dict[str, tuple]] = field(
        init=False, default_factory=lambda: {}
    )
    total_bounding_box_no_pyramid: Optional[Dict[str, tuple]] = field(
        init=False, default_factory=lambda: {}
    )
    verbose: bool = False

    # TODO Is this really needed as a separate class or better integrate directly into CziMetadata class?

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading BoundingBoxes from CZI image data.")

        pyczi_readertype = pyczi.ReaderFileInputTypes.Standard

        if isinstance(self.czisource, Box):
            self.czisource = self.czisource.filepath

        if validators.url(str(self.czisource)):
            pyczi_readertype = pyczi.ReaderFileInputTypes.Curl
            logger.info(
                "FilePath is a valid link. Only pylibCZIrw functionality is available."
            )

        with pyczi.open_czi(str(self.czisource), pyczi_readertype) as czidoc:

            # get scenes bounding rectangles
            try:
                self.scenes_bounding_rect = czidoc.scenes_bounding_rectangle
            except Exception as e:
                if self.verbose:
                    self.scenes_bounding_rect = None
                    logger.info("Scenes Bounding Rectangle not found.")

            try:
                self.scenes_bounding_rect_no_pyramid = (
                    czidoc.scenes_bounding_rectangle_no_pyramid
                )
            except Exception as e:
                if self.verbose:
                    self.scenes_bounding_rect_no_pyramid = None
                    logger.info("Scenes Bounding Rectangle no Pyramid not found.")

            # get total bounding rectangles
            try:
                self.total_rect = czidoc.total_bounding_rectangle
            except Exception as e:
                self.total_rect = None
                if self.verbose:
                    logger.info("Total Bounding Rectangle not found.")

            # get total bounding rectangles
            try:
                self.total_rect_no_pyramid = czidoc.total_bounding_rectangle_no_pyramid
            except Exception as e:
                self.total_rect_no_pyramid = None
                if self.verbose:
                    logger.info("Total Bounding Rectangle no pyramid not found.")

            # get total bounding boxes
            try:
                self.total_bounding_box = czidoc.total_bounding_box
            except Exception as e:
                self.total_bounding_box = None
                if self.verbose:
                    logger.info("Total Bounding Box not found.")

            try:
                self.total_bounding_box_no_pyramid = (
                    czidoc.total_bounding_box_no_pyramid
                )
            except Exception as e:
                self.total_bounding_box_no_pyramid = None
                if self.verbose:
                    logger.info("Total Bounding Box no Pyramid not found.")



# -*- coding: utf-8 -*-

#################################################################
# File        : channel.py
# Author      : sebi06
#
#################################################################

from typing import Union, List, Dict
from dataclasses import dataclass, field
from box import Box, BoxList
import os
from czitools.utils import logging_tools, pixels
from czitools.utils.box import get_czimd_box
from pylibCZIrw import czi as pyczi
import numpy as np

logger = logging_tools.set_logging()


@dataclass
class CziChannelInfo:
    """
    A class to handle channel information from CZI image data.

    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): The source of the CZI image data.
        names (List[str]): List of channel names.
        dyes (List[str]): List of dye names.
        dyes_short (List[str]): List of short dye names.
        channel_descriptions (List[str]): List of channel descriptions.
        colors (List[str]): List of channel colors.
        clims (List[List[float]]): List of channel intensity limits.
        gamma (List[float]): List of gamma values for each channel.
        pixeltypes (Dict[int, str]): Dictionary of pixel types for each channel.
        isRGB (Dict[int, bool]): Dictionary indicating if each channel is RGB.
        consistent_pixeltypes (bool): Indicates if pixel types are consistent across channels.
        czi_disp_settings (Dict[int, pyczi.ChannelDisplaySettingsDataClass]): Dictionary containing the display settings for each channel.
        verbose (bool): Flag to enable verbose logging. Initialized to False.

    Methods:
        __post_init__():
            Initializes the channel information from the CZI image data.
        _get_channel_info(display: Box):
            Extracts and appends channel display information.
        _calculate_display_settings() -> Dict:
            Calculates and returns the display settings for each channel.
    """

    czisource: Union[str, os.PathLike[str], Box]
    names: List[str] = field(init=False, default_factory=lambda: [])
    dyes: List[str] = field(init=False, default_factory=lambda: [])
    dyes_short: List[str] = field(init=False, default_factory=lambda: [])
    channel_descriptions: List[str] = field(init=False, default_factory=lambda: [])
    colors: List[str] = field(init=False, default_factory=lambda: [])
    clims: List[List[float]] = field(init=False, default_factory=lambda: [])
    gamma: List[float] = field(init=False, default_factory=lambda: [])
    pixeltypes: Dict[int, str] = field(init=False, default_factory=lambda: {})
    isRGB: Dict[int, bool] = field(init=False, default_factory=lambda: {})
    consistent_pixeltypes: bool = field(init=False, default=None)
    czi_disp_settings: Dict[int, pyczi.ChannelDisplaySettingsDataClass] = field(
        init=False, default_factory=lambda: {}
    )
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading Channel Information from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        # get channels part of dict
        if czi_box.has_channels:
            try:
                # extract the relevant dimension metadata_tools
                channels = (
                    czi_box.ImageDocument.Metadata.Information.Image.Dimensions.Channels.Channel
                )
                if isinstance(channels, Box):
                    # get the data in case of only one channel
                    (
                        self.names.append("CH1")
                        if channels.Name is None
                        else self.names.append(channels.Name)
                    )
                elif isinstance(channels, BoxList):
                    # get the data in case multiple channels
                    for ch in range(len(channels)):
                        (
                            self.names.append("CH1")
                            if channels[ch].Name is None
                            else self.names.append(channels[ch].Name)
                        )

                # get the pixel types for all channels
                with pyczi.open_czi(
                    str(czi_box.filepath), czi_box.czi_open_arg
                ) as czidoc:

                    # get the pixel typed for all channels
                    self.pixeltypes = czidoc.pixel_types
                    self.isRGB, self.consistent_pixeltypes = pixels.check_if_rgb(
                        self.pixeltypes
                    )

            except AttributeError:
                channels = None
        elif not czi_box.has_channels:
            if self.verbose:
                logger.info("Channel(s) information not found.")

        if czi_box.has_disp:
            try:
                # extract the relevant dimension metadata_tools
                disp = czi_box.ImageDocument.Metadata.DisplaySetting.Channels.Channel
                if isinstance(disp, Box):
                    self._get_channel_info(disp)
                elif isinstance(disp, BoxList):
                    for ch in range(len(disp)):
                        self._get_channel_info(disp[ch])
            except AttributeError:
                disp = None

            # calculate the ChannelDisplaySetting
            self.czi_disp_settings = self._calculate_display_settings()

        elif not czi_box.has_disp:
            if self.verbose:
                logger.info("DisplaySetting(s) not inside CZI metadata.")

    def _get_channel_info(self, display: Box):
        if display is not None:
            (
                self.dyes.append("Dye-CH1")
                if display.Name is None
                else self.dyes.append(display.Name)
            )
            (
                self.dyes_short.append("Dye-CH1")
                if display.ShortName is None
                else self.dyes_short.append(display.ShortName)
            )
            (
                self.channel_descriptions.append("")
                if display.ShortName is None
                else self.channel_descriptions.append(display.Description)
            )
            (
                self.colors.append("#80808000")
                if display.Color is None
                else self.colors.append(display.Color)
            )

            low = 0.0 if display.Low is None else float(display.Low)
            high = 0.5 if display.High is None else float(display.High)

            self.clims.append([low, high])
            (
                self.gamma.append(0.85)
                if display.Gamma is None
                else self.gamma.append(float(display.Gamma))
            )
        else:
            self.dyes.append("Dye-CH1")
            self.colors.append("#80808000")
            self.clims.append([0.0, 0.5])
            self.gamma.append(0.85)

    def _calculate_display_settings(self) -> Dict:

        # get the number of channels
        num_channels = len(self.names)

        # initialize the display settings
        display_settings_dict = {}

        for channel_index in range(num_channels):

            # Get RGB values based on a RGB hexstring. Inside a CZI per channel one gets #AARRGGBB.
            r, g, b = hex_to_rgb(self.colors[channel_index][3:])

            if self.isRGB[channel_index]:
                tinting_mode = pyczi.TintingMode.none
                black_point = 0.0
                white_point = 1.0

            else:
                tinting_mode = pyczi.TintingMode.Color
                black_point = self.clims[channel_index][0]
                white_point = self.clims[channel_index][1]

            display_settings_dict[channel_index] = (
                pyczi.ChannelDisplaySettingsDataClass(
                    is_enabled=True,
                    tinting_mode=tinting_mode,
                    tinting_color=pyczi.Rgb8Color(
                        np.uint8(r), np.uint8(g), np.uint8(b)
                    ),
                    black_point=black_point,  # min value for histogram
                    white_point=white_point,  # max value for histogram
                )
            )

        return display_settings_dict


def hex_to_rgb(hex_string: str) -> tuple[int, int, int]:
    """
    Convert a hexadecimal color string to an RGB tuple.
    Args:
        hex_string (str): A string representing a color in hexadecimal format (e.g., '#RRGGBB').
    Returns:
        tuple: A tuple containing the RGB values (r, g, b) as integers.
    """

    # Remove the '#' characters
    hex_string = hex_string.replace("#", "")

    # check the length of the string
    if len(hex_string) != 6:
        # remove alpha values
        hex_string = hex_string[2:]

    try:
        # Convert hex string to integer values
        r = int(hex_string[0:2], 16)
        g = int(hex_string[2:4], 16)
        b = int(hex_string[4:6], 16)

    except ValueError:

        # Set RGB values to 128 if conversion fails
        r = 128
        g = 128
        b = 128

        logger.error(
            f"Invalid RGB values detected. Conversion of hex string {hex_string} to RGB failed. Setting RGB values to 128."
        )

    return r, g, b



# -*- coding: utf-8 -*-

#################################################################
# File        : czi_metadata.py
# Author      : sebi06
#
# Disclaimer: The code is purely experimental. Feel free to
# use it at your own risk.
#
#################################################################

# from __future__ import annotations
from typing import List, Dict, Tuple, Optional, Any, Union
import os
import xml.etree.ElementTree as ET
from pylibCZIrw import czi as pyczi
from czitools.utils import logging_tools, misc, pixels

# import numpy as np
from dataclasses import dataclass, field
from pathlib import Path
from box import Box
import validators
from dataclasses import asdict
from czitools.metadata_tools.dimension import CziDimensions
from czitools.metadata_tools.boundingbox import CziBoundingBox
from czitools.metadata_tools.attachment import CziAttachments
from czitools.metadata_tools.channel import CziChannelInfo
from czitools.metadata_tools.scaling import CziScaling
from czitools.metadata_tools.sample import CziSampleInfo
from czitools.metadata_tools.objective import CziObjectives
from czitools.metadata_tools.microscope import CziMicroscope
from czitools.metadata_tools.add_metadata import CziAddMetaData
from czitools.metadata_tools.detector import CziDetector
from czitools.utils.box import get_czimd_box
from czitools.metadata_tools.helper import DictObj

logger = logging_tools.set_logging()


@dataclass
class CziMetadata:
    """
    CziMetadata class for handling metadata of CZI image files.

    Attributes:
        filepath (Union[str, os.PathLike[str]]): Path to the CZI image file.
        filename (Optional[str]): Name of the CZI image file.
        dirname (Optional[str]): Directory of the CZI image file.
        is_url (Optional[bool]): Indicates if the filepath is a URL.
        software_name (Optional[str]): Name of the software used for acquisition.
        software_version (Optional[str]): Version of the software used for acquisition.
        acquisition_date (Optional[str]): Date and time of image acquisition.
        creation_date (Optional[str]): Date and time of image creation.
        user_name (Optional[str]): Name of the user who created the image.
        czi_box (Optional[Box]): Metadata box of the CZI file.
        pyczi_dims (Optional[Dict[str, tuple]]): Dimensions of the CZI file.
        aics_dimstring (Optional[str]): Dimension string from aicspylibczi.
        aics_dims_shape (Optional[List[Dict[str, tuple]]]): Dimension shapes from aicspylibczi.
        aics_size (Optional[Tuple[int]]): Size of the CZI file from aicspylibczi.
        aics_ismosaic (Optional[bool]): Indicates if the CZI file is a mosaic.
        aics_dim_order (Optional[Dict[str, int]]): Dimension order from aicspylibczi.
        aics_dim_index (Optional[List[int]]): Dimension indices from aicspylibczi.
        aics_dim_valid (Optional[int]): Number of valid dimensions from aicspylibczi.
        aics_posC (Optional[int]): Position of the 'C' dimension from aicspylibczi.
        pixeltypes (Optional[Dict[int, str]]): Pixel types for all channels.
        consistent_pixeltypes (Optional[bool]): Indicates if pixel types are consistent across channels.
        isRGB (Optional[Dict[int, bool]]): Indicates if the image is RGB.
        has_scenes (Optional[bool]): Indicates if the CZI file has scenes.
        has_label (Optional[bool]): Indicates if the CZI file has a label image.
        has_preview (Optional[bool]): Indicates if the CZI file has a preview image.
        attachments (Optional[CziAttachments]): Attachments in the CZI file.
        npdtype_list (Optional[List[Any]]): Numpy data types for pixel values.
        maxvalue_list (Optional[List[int]]): Maximum values for pixel types.
        image (Optional[CziDimensions]): Dimensions of the CZI image.
        bbox (Optional[CziBoundingBox]): Bounding box of the CZI image.
        channelinfo (Optional[CziChannelInfo]): Information about channels.
        scale (Optional[CziScaling]): Scaling information.
        objective (Optional[CziObjectives]): Objective information.
        detector (Optional[CziDetector]): Detector information.
        microscope (Optional[CziMicroscope]): Microscope information.
        sample (Optional[CziSampleInfo]): Sample information.
        add_metadata (Optional[CziAddMetaData]): Additional metadata.
        scene_size_consistent (Optional[Tuple[int]]): Consistency of scene sizes.
        verbose (bool): Verbose output for logging.

    Methods:
        __post_init__(): Initializes the CziMetadata object after dataclass initialization.
    """

    filepath: Union[str, os.PathLike[str]]
    filename: Optional[str] = field(init=False, default=None)
    dirname: Optional[str] = field(init=False, default=None)
    is_url: Optional[bool] = field(init=False, default=False)
    software_name: Optional[str] = field(init=False, default=None)
    software_version: Optional[str] = field(init=False, default=None)
    acquisition_date: Optional[str] = field(init=False, default=None)
    creation_date: Optional[str] = field(init=False, default=None)
    user_name: Optional[str] = field(init=False, default=None)
    czi_box: Optional[Box] = field(init=False, default=None)
    pyczi_dims: Optional[Dict[str, tuple]] = field(init=False, default_factory=lambda: {})
    aics_dimstring: Optional[str] = field(init=False, default=None)
    aics_dims_shape: Optional[List[Dict[str, tuple]]] = field(init=False, default_factory=lambda: {})
    aics_size: Optional[Tuple[int]] = field(init=False, default_factory=lambda: ())
    aics_ismosaic: Optional[bool] = field(init=False, default=None)
    aics_dim_order: Optional[Dict[str, int]] = field(init=False, default_factory=lambda: {})
    aics_dim_index: Optional[List[int]] = field(init=False, default_factory=lambda: [])
    aics_dim_valid: Optional[int] = field(init=False, default=None)
    aics_posC: Optional[int] = field(init=False, default=None)
    pixeltypes: Optional[Dict[int, str]] = field(init=False, default_factory=lambda: {})
    consistent_pixeltypes: Optional[bool] = field(init=False, default=None)
    isRGB: Optional[Dict[int, bool]] = field(init=False, default_factory=lambda: {})
    has_scenes: Optional[bool] = field(init=False, default=False)
    has_label: Optional[bool] = field(init=False, default=False)
    has_preview: Optional[bool] = field(init=False, default=False)
    attachments: Optional[CziAttachments] = field(init=False, default=None)
    npdtype_list: Optional[List[Any]] = field(init=False, default_factory=lambda: [])
    maxvalue_list: Optional[List[int]] = field(init=False, default_factory=lambda: [])
    image: Optional[CziDimensions] = field(init=False, default=None)
    bbox: Optional[CziBoundingBox] = field(init=False, default=None)
    channelinfo: Optional[CziChannelInfo] = field(init=False, default=None)
    scale: Optional[CziScaling] = field(init=False, default=None)
    objective: Optional[CziObjectives] = field(init=False, default=None)
    detector: Optional[CziDetector] = field(init=False, default=None)
    microscope: Optional[CziMicroscope] = field(init=False, default=None)
    sample: Optional[CziSampleInfo] = field(init=False, default=None)
    add_metadata: Optional[CziAddMetaData] = field(init=False, default=None)
    scene_size_consistent: Optional[Tuple[int]] = field(init=False, default_factory=lambda: ())
    verbose: bool = False

    def __post_init__(self):
        if validators.url(str(self.filepath)):
            self.pyczi_readertype = pyczi.ReaderFileInputTypes.Curl
            self.is_url = True
            if self.verbose:
                logger.info("FilePath is a valid link. Only pylibCZIrw functionality is available.")
        else:
            self.pyczi_readertype = pyczi.ReaderFileInputTypes.Standard
            self.is_url = False

            if isinstance(self.filepath, Path):
                # convert to string
                self.filepath = str(self.filepath)

        # get directory and filename etc.
        self.dirname = str(Path(self.filepath).parent)
        self.filename = str(Path(self.filepath).name)

        # get the metadata_tools as box
        self.czi_box = get_czimd_box(self.filepath)

        # check for existence of scenes
        self.has_scenes = self.czi_box.has_scenes

        # get acquisition data and SW version
        if self.czi_box.ImageDocument.Metadata.Information.Application is not None:
            self.software_name = self.czi_box.ImageDocument.Metadata.Information.Application.Name
            self.software_version = self.czi_box.ImageDocument.Metadata.Information.Application.Version

        if self.czi_box.ImageDocument.Metadata.Information.Image is not None:
            self.acquisition_date = self.czi_box.ImageDocument.Metadata.Information.Image.AcquisitionDateAndTime

        if self.czi_box.ImageDocument.Metadata.Information.Document is not None:
            self.creation_date = self.czi_box.ImageDocument.Metadata.Information.Document.CreationDate
            self.user_name = self.czi_box.ImageDocument.Metadata.Information.Document.UserName

        # get the dimensions and order
        self.image = CziDimensions(self.czi_box, verbose=self.verbose)

        # get metadata_tools using pylibCZIrw
        with pyczi.open_czi(self.filepath, self.pyczi_readertype) as czidoc:
            # get dimensions
            self.pyczi_dims = czidoc.total_bounding_box

            # get the pixel typed for all channels
            self.pixeltypes = czidoc.pixel_types
            self.isRGB, self.consistent_pixeltypes = pixels.check_if_rgb(self.pixeltypes)

            # check for consistent scene shape
            self.scene_shape_is_consistent = pixels.check_scenes_shape(czidoc, size_s=self.image.SizeS)

        if not self.is_url:
            # get some additional metadata_tools using aicspylibczi
            try:
                from aicspylibczi import CziFile

                # get the general CZI object using aicspylibczi
                aicsczi = CziFile(self.filepath)

                self.aics_dimstring = aicsczi.dims
                self.aics_dims_shape = aicsczi.get_dims_shape()
                self.aics_size = aicsczi.size
                self.aics_ismosaic = aicsczi.is_mosaic()
                (
                    self.aics_dim_order,
                    self.aics_dim_index,
                    self.aics_dim_valid,
                ) = pixels.get_dimorder(aicsczi.dims)
                self.aics_posC = self.aics_dim_order["C"]

            except ImportError as e:
                # print("Package aicspylibczi not found. Use Fallback values.")
                logger.warning(f" {e} : Package aicspylibczi not found. Use Fallback values.")

        self.npdtype_list = []
        self.maxvalue_list = []

        for ch, px in self.pixeltypes.items():
            npdtype, maxvalue = pixels.get_dtype_fromstring(px)
            self.npdtype_list.append(npdtype)
            self.maxvalue_list.append(maxvalue)

        # try to guess if the CZI is a mosaic file
        if self.image.SizeM is None or self.image.SizeM == 1:
            self.ismosaic = False
        else:
            self.ismosaic = True

        # get the bounding boxes
        self.bbox = CziBoundingBox(self.czi_box, verbose=self.verbose)

        # get information about channels
        self.channelinfo = CziChannelInfo(self.czi_box, verbose=self.verbose)

        # get scaling info
        self.scale = CziScaling(self.czi_box, verbose=self.verbose)

        # get objective information
        self.objective = CziObjectives(self.czi_box, verbose=self.verbose)

        # get detector information
        self.detector = CziDetector(self.czi_box, verbose=self.verbose)

        # get detector information
        self.microscope = CziMicroscope(self.czi_box, verbose=self.verbose)

        # get information about sample carrier and wells etc.
        self.sample = CziSampleInfo(self.czi_box, verbose=self.verbose)

        # get additional metainformation
        self.add_metadata = CziAddMetaData(self.czi_box, verbose=self.verbose)

        # check for attached label or preview image
        self.attachments = CziAttachments(self.czi_box, verbose=self.verbose)


def get_metadata_as_object(filepath: Union[str, os.PathLike[str]]) -> DictObj:
    """
    Get the complete CZI metadata as an object.
    This function reads the metadata from a CZI file and converts it into a
    dictionary, which is then used to create an object of type DictObj.
    Args:
        filepath (Union[str, os.PathLike[str]]): The path to the CZI file.
            This can be a string or an os.PathLike object.
    Returns:
        DictObj: An object containing the metadata extracted from the CZI file.
    """

    if isinstance(filepath, Path):
        # convert to string
        filepath = str(filepath)

    # get metadata_tools dictionary using pylibCZIrw
    with pyczi.open_czi(filepath) as czidoc:
        md_dict = czidoc.metadata

    return DictObj(md_dict)


def obj2dict(obj: Any, sort: bool = True) -> Dict[str, Any]:
    """
    obj2dict: Convert a class attributes and their values to a dictionary

    Args:
        obj (Any): Class to be converted
        sort (bool, optional): Sort the resulting dictionary by key names. Defaults to True.

    Returns:
        Dict[str, Any]: The resulting disctionary.
    """

    # https://stackoverflow.com/questions/7963762/what-is-the-most-economical-way-to-convert-nested-python-objects-to-dictionaries

    if not hasattr(obj, "__dict__"):
        return obj

    result = {}

    for key, val in obj.__dict__.items():
        if key.startswith("_"):
            continue

        element = []

        if isinstance(val, list):
            for item in val:
                element.append(obj2dict(item))
        else:
            element = obj2dict(val)

        result[key] = element

    # delete key "czisource"
    if "czisource" in result.keys():
        del result["czisource"]

    if sort:
        return misc.sort_dict_by_key(result)

    elif not sort:
        return result


def writexml(filepath: Union[str, os.PathLike[str]], xmlsuffix: str = "_CZI_MetaData.xml") -> str:
    """
    writexml: Write XML information of CZI to disk

    Args:
        filepath (Union[str, os.PathLike[str]]): CZI image filename
        xmlsuffix (str, optional): suffix for the XML file that will be created, defaults to '_CZI_MetaData.xml'. Defaults to '_CZI_MetaData.xml'.

    Returns:
        str: filename of the XML file
    """

    if isinstance(filepath, Path):
        # convert to string
        filepath = str(filepath)

    # get the raw metadata_tools as XML or dictionary
    with pyczi.open_czi(filepath) as czidoc:
        metadata_xmlstr = czidoc.raw_metadata

    # change file name
    xmlfile = filepath.replace(".czi", xmlsuffix)

    # get tree from string
    tree = ET.ElementTree(ET.fromstring(metadata_xmlstr))

    # write XML file to same folder
    tree.write(xmlfile, encoding="utf-8", method="xml")

    return xmlfile


def create_md_dict_red(metadata: CziMetadata, sort: bool = True, remove_none: bool = True) -> Dict:
    """
    create_mdict_red: Created a reduced metadata_tools dictionary

    Args:
        metadata: CziMetadata class
        sort: sort the dictionary
        remove_none: Remove None values from dictionary

    Returns: dictionary with the metadata_tools

    """

    # create a dictionary with the metadata_tools
    md_dict = {
        "Directory": metadata.dirname,
        "Filename": metadata.filename,
        "AcqDate": metadata.acquisition_date,
        "CreationDate": metadata.creation_date,
        "UserName": metadata.user_name,
        "SW-App": metadata.software_version,
        "SW-Version": metadata.software_name,
        "SizeX": metadata.image.SizeX,
        "SizeY": metadata.image.SizeY,
        "SizeZ": metadata.image.SizeZ,
        "SizeC": metadata.image.SizeC,
        "SizeT": metadata.image.SizeT,
        "SizeS": metadata.image.SizeS,
        "SizeB": metadata.image.SizeB,
        "SizeM": metadata.image.SizeM,
        "SizeH": metadata.image.SizeH,
        "SizeI": metadata.image.SizeI,
        "isRGB": metadata.isRGB,
        "has_scenes": metadata.has_scenes,
        "has_label": metadata.attachments.has_label,
        "has_preview": metadata.attachments.has_preview,
        "has_prescan": metadata.attachments.has_prescan,
        "ismosaic": metadata.ismosaic,
        "ObjNA": metadata.objective.NA,
        "ObjMag": metadata.objective.totalmag,
        "TubelensMag": metadata.objective.tubelensmag,
        "XScale": metadata.scale.X,
        "YScale": metadata.scale.Y,
        "XScale_sf": metadata.scale.X_sf,
        "YScale_sf": metadata.scale.Y_sf,
        "ZScale": metadata.scale.Z,
        "ScaleRatio_XYZ": metadata.scale.ratio,
        "ChannelNames": metadata.channelinfo.names,
        "ChannelDyes": metadata.channelinfo.dyes,
        "ChannelDyesShort": metadata.channelinfo.dyes_short,
        "ChannelColors": metadata.channelinfo.colors,
        "WellArrayNames": metadata.sample.well_array_names,
        "ChannelDescriptions": metadata.channelinfo.channel_descriptions,
        "WellIndicies": metadata.sample.well_indices,
        "WellPositionNames": metadata.sample.well_position_names,
        "WellRowID": metadata.sample.well_rowID,
        "WellColumnID": metadata.sample.well_colID,
        "WellCounter": metadata.sample.well_counter,
        "SceneCenterStageX": metadata.sample.scene_stageX,
        "SceneCenterStageY": metadata.sample.scene_stageX,
        "ImageStageX": metadata.sample.image_stageX,
        "ImageStageY": metadata.sample.image_stageX,
        "TotalBoundingBox": metadata.bbox.total_bounding_box,
    }

    # check for extra entries when reading mosaic file with a scale factor
    if hasattr(metadata.image, "SizeX_sf"):
        md_dict["XScale_sf"] = metadata.scale.X_sf
        md_dict["YScale_sf"] = metadata.scale.Y_sf

    if metadata.has_scenes:
        md_dict["SizeX_scene"] = metadata.image.SizeX_scene
        md_dict["SizeY_scene"] = metadata.image.SizeY_scene

    if remove_none:
        # md_dict = misc.remove_none_from_dict(md_dict)
        md_dict = misc.clean_dict(md_dict)

    if sort:
        return misc.sort_dict_by_key(md_dict)
    if not sort:
        return md_dict


def create_md_dict_nested(metadata: CziMetadata, sort: bool = True, remove_none: bool = True) -> Dict:
    """
    Create nested dictionary from metadata_tools

    Args:
        metadata (CziMetadata): CzIMetaData object_
        sort (bool, optional): Sort the dictionary_. Defaults to True.
        remove_none (bool, optional): Remove None values from dictionary. Defaults to True.

    Returns:
        Dict: Nested dictionary with reduced set of metadata_tools
    """

    md_box_image = Box(asdict(metadata.image))
    del md_box_image.czisource

    md_box_scale = Box(asdict(metadata.scale))
    del md_box_scale.czisource

    md_box_sample = Box(asdict(metadata.sample))
    del md_box_sample.czisource

    md_box_objective = Box(asdict(metadata.objective))
    del md_box_objective.czisource

    md_box_channels = Box(asdict(metadata.channelinfo))
    del md_box_channels.czisource

    md_box_bbox = Box(metadata.bbox.total_bounding_box)
    # del md_box_bbox.czisource

    md_box_info = Box(
        {
            "Directory": metadata.dirname,
            "Filename": metadata.filename,
            "AcqDate": metadata.acquisition_date,
            "CreationDate": metadata.creation_date,
            "UserName": metadata.user_name,
            "SW-App": metadata.software_version,
            "SW-Version": metadata.software_name,
        }
    )

    md_box_image_add = Box(
        {
            "isRGB": metadata.isRGB,
            "has_scenes": metadata.has_scenes,
            "ismosaic": metadata.ismosaic,
        }
    )

    md_box_image += md_box_image_add

    IDs = [
        "image",
        "scale",
        "sample",
        "objectives",
        "channels",
        "bbox",
        "info",
    ]

    mds = [
        md_box_image.to_dict(),
        md_box_scale.to_dict(),
        md_box_sample.to_dict(),
        md_box_objective.to_dict(),
        md_box_channels.to_dict(),
        md_box_bbox.to_dict(),
        md_box_info.to_dict(),
    ]

    md_dict = dict(zip(IDs, mds))

    if remove_none:
        # md_dict = misc.remove_none_from_dict(md_dict)
        md_dict = misc.clean_dict(md_dict)

    if sort:
        return misc.sort_dict_by_key(md_dict)
    if not sort:
        return md_dict

    return md_dict



from typing import Tuple, Optional, Union, List
from dataclasses import dataclass, field, fields, Field
from box import Box
import os
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box
from pylibCZIrw import czi as pyczi

from pydantic.dataclasses import dataclass
from pydantic import BaseModel, ConfigDict
from bioio import BioImage

logger = logging_tools.set_logging()


def string_to_float_list(string: str) -> List[float]:
    """Converts a space-separated string of numbers into a list of floats.

    Args:
      string: The input string.

    Returns:
      A list of floats.
    """

    numbers = string.split()
    float_numbers = [float(num) for num in numbers]

    return float_numbers


class Config:
    arbitrary_types_allowed = True


@dataclass(config=Config)
class CziDimensions:
    """
    CziDimensions is a dataclass that encapsulates the dimensions of a CZI image.
    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): Source of the CZI image.
        SizeX (Optional[int]): Total size in the X dimension, including scenes.
        SizeY (Optional[int]): Total size in the Y dimension, including scenes.
        SizeX_scene (Optional[int]): Size in the X dimension per scene (if equal scene sizes).
        SizeY_scene (Optional[int]): Size in the Y dimension per scene (if equal scene sizes).
        SizeS (Optional[int]): Size in the S dimension.
        SizeT (Optional[int]): Size in the T dimension.
        SizeZ (Optional[int]): Size in the Z dimension.
        SizeC (Optional[int]): Size in the C dimension.
        SizeM (Optional[int]): Size in the M dimension.
        SizeR (Optional[int]): Size in the R dimension.
        SizeH (Optional[int]): Size in the H dimension.
        SizeI (Optional[int]): Size in the I dimension.
        SizeV (Optional[int]): Size in the V dimension.
        SizeB (Optional[int]): Size in the B dimension.
        posZ (Optional[List[float]]): List of Z positions in microns, if they exist.
        posT (Optional[List[float]]): List of T positions in seconds, if they exist.
        verbose (bool): Flag to enable verbose logging.
    Methods:
        __post_init__(): Initializes the dimensions by reading from the CZI image data.
        set_dimensions(): Populates the image dimensions with the detected values from the metadata.
        set_dimensions_new(): Populates the image dimensions with the detected values from the metadata using a new method.
    Notes:
        The class contains information about official CZI Dimension Characters:
        "X": "Width"        : width of image [pixel]
        "Y": "Height"       : height of image [pixel]
        "C": "Channel"      : number of channels
        "Z": "Slice"        : number of z-planes
        "T": "Time"         : number of time points
        "R": "Rotation"     :
        "S": "Scene"        : contiguous regions of interest in a mosaic image
        "I": "Illumination" : SPIM direction for LightSheet
        "B": "Block"        : acquisition
        "M": "Mosaic"       : index of tile for compositing a scene
        "V": "View"         : e.g. for SPIM
        In addition, it contains the Z-Positions [microns] and T-Positions [s] if they exist. Otherwise, they
        are set to None.
        "H": "Phase"        : e.g. Airy detector fibers
    """

    czisource: Union[str, os.PathLike[str], Box]
    SizeX: Optional[int] = field(
        init=False, default=None
    )  # total size X including scenes
    SizeY: Optional[int] = field(
        init=False, default=None
    )  # total size Y including scenes
    SizeX_scene: Optional[List[int]] = field(
        init=False, default_factory=lambda: []
    )  # size X per scene (if equal scene sizes)
    SizeY_scene: Optional[List[int]] = field(
        init=False, default_factory=lambda: []
    )  # size Y per scene (if equal scene sizes)
    SizeS: Optional[int] = field(init=False, default=None)
    SizeT: Optional[int] = field(init=False, default=None)
    SizeZ: Optional[int] = field(init=False, default=None)
    SizeC: Optional[int] = field(init=False, default=None)
    SizeM: Optional[int] = field(init=False, default=None)
    SizeR: Optional[int] = field(init=False, default=None)
    SizeH: Optional[int] = field(init=False, default=None)
    SizeI: Optional[int] = field(init=False, default=None)
    SizeV: Optional[int] = field(init=False, default=None)
    SizeB: Optional[int] = field(init=False, default=None)
    posZ: Optional[List[float]] = field(init=False, default=None)
    posT: Optional[List[float]] = field(init=False, default=None)
    verbose: bool = False

    def __post_init__(self):

        if self.verbose:
            logger.info("Reading Dimensions from CZI image data.")
        self.set_dimensions()

        # set dimensions in XY with respect to possible down scaling
        self.SizeX_sf = self.SizeX
        self.SizeY_sf = self.SizeY

    def set_dimensions(self):
        """
        Populate the image dimensions with the detected values from the metadata.
        This method sets the dimensions of the image based on the metadata extracted from the CZI file.
        It handles various dimensions such as SizeX, SizeY, SizeS, SizeT, SizeZ, SizeC, SizeM, SizeR, SizeH, SizeI, SizeV, and SizeB.
        Additionally, it processes scene dimensions and time (T) and z-position (Z) lists if available.
        Attributes:
            SizeX (int): Width of the image.
            SizeY (int): Height of the image.
            SizeS (int): Size of the S dimension.
            SizeT (int): Size of the T dimension.
            SizeZ (int): Size of the Z dimension.
            SizeC (int): Size of the C dimension.
            SizeM (int): Size of the M dimension.
            SizeR (int): Size of the R dimension.
            SizeH (int): Size of the H dimension.
            SizeI (int): Size of the I dimension.
            SizeV (int): Size of the V dimension.
            SizeB (int): Size of the B dimension.
            SizeX_scene (int): Width of the scene bounding rectangle.
            SizeY_scene (int): Height of the scene bounding rectangle.
            posT (list of float): List of time positions.
            posZ (list of float): List of z positions.
        Raises:
            KeyError: If scene bounding rectangle information is not found.
            Exception: If there is an error converting string to a float list for posT or posZ.
        """

        # get the Box and extract the relevant dimension metadata_tools
        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        dimensions = czi_box.ImageDocument.Metadata.Information.Image

        # define the image dimensions to check for
        dims = [
            "SizeX",
            "SizeY",
            "SizeS",
            "SizeT",
            "SizeZ",
            "SizeC",
            "SizeM",
            "SizeR",
            "SizeH",
            "SizeI",
            "SizeV",
            "SizeB",
        ]

        cls_fields: Tuple[Field, ...] = fields(self.__class__)
        for fd in cls_fields:
            if fd.name in dims:
                if dimensions[fd.name] is not None:
                    setattr(self, fd.name, int(dimensions[fd.name]))

        if czi_box.has_scenes:
            try:
                with pyczi.open_czi(czi_box.filepath, czi_box.czi_open_arg) as czidoc:
                    self.SizeX_scene = czidoc.scenes_bounding_rectangle_no_pyramid[0].w
                    self.SizeY_scene = czidoc.scenes_bounding_rectangle_no_pyramid[0].h
            except KeyError as e:
                self.SizeX_scene = None
                self.SizeY_scene = None
                if self.verbose:
                    logger.warning(
                        "Scenes Dimension detected but no bounding rectangle information found."
                    )

        if czi_box.has_T:
            # check if there is a list with timepoints (is not in very CZI)
            if dimensions.Dimensions.T.Positions is not None:
                if dimensions.Dimensions.T.Positions.List is not None:
                    try:
                        self.posT = string_to_float_list(
                            dimensions.Dimensions.T.Positions.List.Offsets
                        )
                    except Exception as e:
                        if self.verbose:
                            logger.error(f"{e}")
                else:
                    if self.verbose:
                        logger.warning(
                            "No posT list found under 'dimensions.Dimensions.T.Positions.List'"
                        )
            else:
                if self.verbose:
                    logger.warning(
                        "No posT list found under 'dimensions.Dimensions.T.Positions'"
                    )

        if czi_box.has_Z:
            # check if there is a list with z-positions (is not in very CZI)
            if dimensions.Dimensions.Z.Positions is not None:
                if dimensions.Dimensions.Z.Positions.List is not None:
                    try:
                        self.posZ = string_to_float_list(
                            dimensions.Dimensions.Z.Positions.List.Offsets
                        )
                    except Exception as e:
                        if self.verbose:
                            logger.error(f"{e}")
                else:
                    if self.verbose:
                        logger.warning(
                            "No posZ list found under 'dimensions.Dimensions.Z.Positions.List'"
                        )
            else:
                if self.verbose:
                    logger.warning(
                        "No posZ list found under 'dimensions.Dimensions.Z.Positions'"
                    )

    def set_dimensions_2(self):
        """
        Set the dimensions of the CZI file based on metadata and subblock information.
        This method extracts dimension metadata from the provided CZI source and calculates
        the sizes of various dimensions (e.g., X, Y, T, Z, etc.) using the total bounding box
        of the CZI file. It also attempts to retrieve positional information for time (T) and
        Z-dimensions if available.

        Attributes Set:
            - SizeX, SizeY, SizeT, SizeZ, SizeC, SizeR, SizeH, SizeI, SizeV, SizeB:
              Sizes of respective dimensions if present in the total bounding box.
            - SizeS: Number of scenes in the CZI file.
            - SizeM: Number of mosaic tiles, determined using BioImage or metadata.
            - SizeX_scene, SizeY_scene: Scene dimensions if bounding rectangle information is available.
            - posT: List of time positions if available in the metadata.
            - posZ: List of Z positions if available in the metadata.

        Notes:
            - The method uses `pyczi.open_czi` to open the CZI file and extract subblock information.
            - Positional information for T and Z dimensions is optional and may not be present in all CZI files.
            - Verbose logging is used to provide detailed warnings or errors during execution.

        Raises:
            - Exception: If an error occurs while parsing positional information for T or Z dimensions.
        """

        # get the Box and extract the relevant dimension metadata_tools
        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        dimensions = czi_box.ImageDocument.Metadata.Information.Image

        with pyczi.open_czi(self.czisource, czi_box.czi_open_arg) as czidoc:

            # get the sizes from the subblocks and not from the metadata
            tb = czidoc.total_bounding_box_no_pyramid

            for dim in ["X", "Y", "T", "Z", "C", "R", "H", "I", "V", "B"]:
                if dim in tb.keys():
                    setattr(self, f"Size{dim}", tb[dim][1])

            self.SizeS = len(czidoc.scenes_bounding_rectangle_no_pyramid)

            if czi_box.czi_open_arg == pyczi.ReaderFileInputTypes.Standard:

                # try to read the number of Mosaic tiles in using bioio-czi and aicspylibczi
                img = BioImage(
                    self.czisource,
                    reconstruct_mosaic=False,
                    include_subblock_metadata=True,
                    use_aicspylibczi=True,
                )

                if hasattr(img.dims, "M"):
                    self.SizeM = img.dims.M

            elif czi_box.czi_open_arg == pyczi.ReaderFileInputTypes.Curl:

                if dimensions.SizeM is not None:
                    # try to read the number of Mosaic tiles from the metadata XML
                    self.SizeM = dimensions.SizeM

            if self.SizeS is not None:

                for s in range(len(czidoc.scenes_bounding_rectangle_no_pyramid)):
                    if czidoc.scenes_bounding_rectangle_no_pyramid[s].w is not None:
                        self.SizeX_scene.append(
                            czidoc.scenes_bounding_rectangle_no_pyramid[s].w
                        )
                    if czidoc.scenes_bounding_rectangle_no_pyramid[s].h is not None:
                        self.SizeY_scene.append(
                            czidoc.scenes_bounding_rectangle_no_pyramid[s].h
                        )

        if czi_box.has_T:
            # check if there is a list with timepoints (is not in very CZI)
            if dimensions.Dimensions.T.Positions is not None:
                if dimensions.Dimensions.T.Positions.List is not None:
                    try:
                        self.posT = string_to_float_list(
                            dimensions.Dimensions.T.Positions.List.Offsets
                        )
                    except Exception as e:
                        if self.verbose:
                            logger.error(f"{e}")
                else:
                    if self.verbose:
                        logger.warning(
                            "No posT list found under 'dimensions.Dimensions.T.Positions.List'"
                        )
            else:
                if self.verbose:
                    logger.warning(
                        "No posT list found under 'dimensions.Dimensions.T.Positions'"
                    )

        if czi_box.has_Z:
            # check if there is a list with z-positions (is not in very CZI)
            if dimensions.Dimensions.Z.Positions is not None:
                if dimensions.Dimensions.Z.Positions.List is not None:
                    try:
                        self.posZ = string_to_float_list(
                            dimensions.Dimensions.Z.Positions.List.Offsets
                        )
                    except Exception as e:
                        if self.verbose:
                            logger.error(f"{e}")
                else:
                    if self.verbose:
                        logger.warning(
                            "No posZ list found under 'dimensions.Dimensions.Z.Positions.List'"
                        )
            else:
                if self.verbose:
                    logger.warning(
                        "No posZ list found under 'dimensions.Dimensions.Z.Positions'"
                    )



import sys
import logging

try:
    import colorlog

    HAS_COLORLOG = True
except ImportError:
    HAS_COLORLOG = False


def set_logging(level=logging.INFO, format_string=None, colorize=True):
    """
    Configures a standard Python logger with customizable level and format.

    Args:
        level: Logging level (e.g., logging.INFO, logging.DEBUG, logging.WARNING)
        format_string: Custom format string for log messages. If None, uses default format.
        colorize: Whether to colorize the output (requires colorlog package)

    Returns:
        logging.Logger: Configured standard Python logger instance.
    """

    # Create a logger
    logger = logging.getLogger("czitools")

    # Clear any existing handlers
    logger.handlers.clear()

    # Set the logging level
    logger.setLevel(level)

    # Create console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(level)

    # Create formatter
    if colorize and HAS_COLORLOG:
        # Use colorlog for colored output
        if format_string is None:
            format_string = "%(log_color)s%(asctime)s - %(name)s - %(levelname)s - %(message)s"

        formatter = colorlog.ColoredFormatter(
            format_string,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
    else:
        # Use standard formatter
        if format_string is None:
            format_string = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

        formatter = logging.Formatter(format_string)

        # If colorlog is not available but colorize was requested, warn the user
        if colorize and not HAS_COLORLOG:
            print("Warning: colorlog package not found. Install with 'pip install colorlog' for colored output.")

    console_handler.setFormatter(formatter)

    # Add handler to logger
    logger.addHandler(console_handler)

    return logger



from typing import Union, List, Dict, Tuple
from dataclasses import dataclass, field
from box import Box, BoxList
import os
import numpy as np
from collections import Counter
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box
from czitools.utils.misc import get_planetable
from czitools.metadata_tools.dimension import CziDimensions
import traceback

logger = logging_tools.set_logging()


@dataclass
class CziSampleInfo:
    """
    A class to represent and extract sample information from CZI image data.
    Attributes:
    -----------
    czisource : Union[str, os.PathLike[str], Box]
        The source of the CZI image data.
    well_array_names : List[str]
        Names of the well arrays.
    well_indices : List[int]
        Indices of the wells.
    well_position_names : List[str]
        Position names of the wells.
    well_colID : List[int]
        Column IDs of the wells.
    well_rowID : List[int]
        Row IDs of the wells.
    well_counter : Dict
        Counter for the well instances.
    scene_stageX : List[float]
        X coordinates of the scene stages.
    scene_stageY : List[float]
        Y coordinates of the scene stages.
    image_stageX : float
        X coordinate of the image stage.
    image_stageY : float
        Y coordinate of the image stage.
    multipos_per_well : bool
        Flag to indicate multiple positions per well.
    verbose (bool):
        Flag to enable verbose logging.
    Methods:
    --------
    __post_init__():
        Initializes the CziSampleInfo object and reads sample carrier information from CZI image data.
    get_well_info(well: Box):
        Extracts well information from the given well Box object.
    """

    czisource: Union[str, os.PathLike[str], Box]
    well_array_names: List[str] = field(init=False, default_factory=lambda: [])
    well_indices: List[int] = field(init=False, default_factory=lambda: [])
    well_position_names: List[str] = field(init=False, default_factory=lambda: [])
    well_colID: List[int] = field(init=False, default_factory=lambda: [])
    well_rowID: List[int] = field(init=False, default_factory=lambda: [])
    well_counter: Dict = field(init=False, default_factory=lambda: {})
    scene_stageX: List[float] = field(init=False, default_factory=lambda: [])
    scene_stageY: List[float] = field(init=False, default_factory=lambda: [])
    image_stageX: float = field(init=False, default=None)
    image_stageY: float = field(init=False, default=None)
    multipos_per_well: bool = False
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading SampleCarrier Information from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        size_s = CziDimensions(czi_box).SizeS

        if size_s is not None:
            try:
                allscenes = czi_box.ImageDocument.Metadata.Information.Image.Dimensions.S.Scenes.Scene

                # check if there are multiple positions per well
                # self.multipos_per_well = check_multipos_well(allscenes[0])

                if isinstance(allscenes, Box):
                    # check if there are multiple positions per well
                    self.multipos_per_well = check_multipos_well(allscenes)
                    self.get_well_info(allscenes)

                if isinstance(allscenes, BoxList):
                    # check if there are multiple positions per well
                    self.multipos_per_well = check_multipos_well(allscenes[0])
                    for well in range(len(allscenes)):
                        self.get_well_info(allscenes[well])

            except AttributeError:
                if self.verbose:
                    logger.info("CZI contains no scene metadata_tools.")

        elif size_s is None:
            if self.verbose:
                logger.info("No Scene or Well information found. Try to read XY Stage Coordinates from subblocks.")

            try:
                # read the data from CSV file from a single plane
                planetable = get_planetable(czi_box.filepath, scene=0, tile=0, time=0, channel=0, zplane=0)

                self.image_stageX = float(planetable["X[micron]"][0])
                self.image_stageY = float(planetable["Y[micron]"][0])

            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    logger.error(e)

    def get_well_info(self, well: Box):
        """
        Extracts and processes well information from a given Box object.
        Parameters:
        well (Box): A Box object containing well information.
        The method performs the following actions:
        - Appends the well's ArrayName to self.well_array_names and updates self.well_counter.
        - Appends the well's Index to self.well_indices. If Index is None, logs a message and appends 1.
        - Appends the well's Name to self.well_position_names. If Name is None, logs a message and appends "P1".
        - Appends the well's Shape ColumnIndex and RowIndex to self.well_colID and self.well_rowID respectively. If Shape is None, logs a message and appends 0 to both.
        - Appends the well's CenterPosition coordinates to self.scene_stageX and self.scene_stageY. If CenterPosition is None, logs a message and appends 0.0 to both.
        """

        # check the ArrayName
        if well.ArrayName is not None:
            self.well_array_names.append(well.ArrayName)
            # count the well instances
            self.well_counter = Counter(self.well_array_names)

        if well.Index is not None:
            self.well_indices.append(int(well.Index))
        elif well.Index is None:
            if self.verbose:
                logger.info("Well Index not found.")
                self.well_indices.append(1)

        if well.Name is not None:
            self.well_position_names.append(well.Name)
        elif well.Name is None:
            if self.verbose:
                logger.info("Well Position Names not found.")
                self.well_position_names.append("P1")

        if well.Shape is not None:
            self.well_colID.append(int(well.Shape.ColumnIndex))
            self.well_rowID.append(int(well.Shape.RowIndex))
        elif well.Shape is None:
            if self.verbose:
                logger.info("Well Column or Row IDs not found.")
                self.well_colID.append(0)
                self.well_rowID.append(0)

        if well.CenterPosition is not None:
            # get the SceneCenter Position
            sx = well.CenterPosition.split(",")[0]
            sy = well.CenterPosition.split(",")[1]
            self.scene_stageX.append(np.double(sx))
            self.scene_stageY.append(np.double(sy))
        elif well.CenterPosition is None:
            if self.verbose:
                logger.info("Stage Positions XY not found.")
                self.scene_stageX.append(0.0)
                self.scene_stageY.append(0.0)


def get_scenes_for_well(sample: CziSampleInfo, wellID: str) -> Tuple[int]:
    """
    Returns a list of scene indices for a given well ID. The list of wells is stored inside
    the "sample" section of the metadata

    Args:
        sample_info (CziSampleInfo): The CziSampleInfo object containing well information.
        wellID (str): The ID of the well.

    Returns:
        Tuple[int]: A list of scene indices corresponding to the given well ID.
    """

    if sample.multipos_per_well:
        scene_indices = [i for i, x in enumerate(sample.well_array_names) if x == wellID]
    if not sample.multipos_per_well:
        scene_indices = [i for i, x in enumerate(sample.well_position_names) if x == wellID]

    return tuple(scene_indices)


def check_multipos_well(scene: Box):

    if scene.ArrayName is None and scene.Name is not None:
        return False
    else:
        return True



from typing import Optional, Union, List
from dataclasses import dataclass, field
from box import Box
import os
from czitools.utils import logging_tools
from pathlib import Path
import validators

logger = logging_tools.set_logging()


@dataclass
class CziAttachments:
    """
    CziAttachments class for handling CZI image data attachments.
    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): Source of the CZI image data.
        has_label (Optional[bool]): Indicates if the CZI image has a label attachment.
        has_preview (Optional[bool]): Indicates if the CZI image has a preview attachment.
        has_prescan (Optional[bool]): Indicates if the CZI image has a prescan attachment.
        names (Optional[List[str]]): List of attachment names found in the CZI image.
        verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods:
        __post_init__(): Initializes the CziAttachments object, reads attachment images from the CZI image data, and sets the appropriate flags for label, preview, and prescan attachments.
    """

    czisource: Union[str, os.PathLike[str], Box]
    has_label: Optional[bool] = field(init=False, default=False)
    has_preview: Optional[bool] = field(init=False, default=False)
    has_prescan: Optional[bool] = field(init=False, default=False)
    names: Optional[List[str]] = field(init=False, default_factory=lambda: [])
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading AttachmentImages from CZI image data.")

        try:
            import czifile

            if isinstance(self.czisource, Path):
                # convert to string
                self.czisource = str(self.czisource)
            elif isinstance(self.czisource, Box):
                self.czisource = self.czisource.filepath

            if validators.url(self.czisource):
                if self.verbose:
                    logger.warning(
                        "Reading Attachments from CZI via a link is not supported."
                    )
            else:
                # create CZI-object using czifile library
                with czifile.CziFile(self.czisource) as cz:
                    # iterate over attachments
                    for att in cz.attachments():
                        self.names.append(att.attachment_entry.name)

                    if "SlidePreview" in self.names:
                        self.has_preview = True
                        if self.verbose:
                            logger.info("Attachment SlidePreview found.")
                    if "Label" in self.names:
                        self.has_label = True
                        if self.verbose:
                            logger.info("Attachment Label found.")
                    if "Prescan" in self.names:
                        self.has_prescan = True
                        if self.verbose:
                            logger.info("Attachment Prescan found.")

        except ImportError as e:
            if self.verbose:
                logger.warning(
                    f"{e}: Package czifile not found. Cannot extract information about attached images."
                )



from typing import Optional, Union
from dataclasses import dataclass, field
import os
from czitools.utils import logging_tools
from pylibCZIrw import czi as pyczi
from pathlib import Path

logger = logging_tools.set_logging()


@dataclass
class CziScene:
    """
    CziScene class represents a scene within a CZI (Carl Zeiss Image) file.
    Attributes:
        filepath (Union[str, os.PathLike[str]]): The path to the CZI file.
        index (int): The index of the scene within the CZI file.
        bbox (Optional[pyczi.Rectangle]): The bounding box of the scene. Initialized to None.
        xstart (Optional[int]): The starting x-coordinate of the scene. Initialized to None.
        ystart (Optional[int]): The starting y-coordinate of the scene. Initialized to None.
        width (Optional[int]): The width of the scene. Initialized to None.
        height (Optional[int]): The height of the scene. Initialized to None.
        verbose (bool): Flag to enable verbose logging.
    Methods:
        __post_init__(): Initializes the scene information by reading from the CZI file.
    """

    filepath: Union[str, os.PathLike[str]]
    index: int
    bbox: Optional[pyczi.Rectangle] = field(init=False, default=None)
    xstart: Optional[int] = field(init=False, default=None)
    ystart: Optional[int] = field(init=False, default=None)
    width: Optional[int] = field(init=False, default=None)
    height: Optional[int] = field(init=False, default=None)
    verbose: bool = False

    def __post_init__(self):
        logger.info("Reading Scene Information from CZI image data.")

        if isinstance(self.filepath, Path):
            # convert to string
            self.filepath = str(self.filepath)

        # get scene information from the CZI file
        with pyczi.open_czi(self.filepath) as czidoc:
            try:
                self.bbox = czidoc.scenes_bounding_rectangle[self.index]
                self.xstart = self.bbox.x
                self.ystart = self.bbox.y
                self.width = self.bbox.w
                self.height = self.bbox.h
            except KeyError as e:
                if self.verbose:
                    # in case an invalid index was used
                    logger.info(f"{e}: No Scenes detected.")



from typing import Union, List
from dataclasses import dataclass, field
from box import Box, BoxList
import os
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box

logger = logging_tools.set_logging()


@dataclass
class CziDetector:
    """
    A class to represent and extract detector information from CZI image data.
    Attributes
    ----------
    czisource : Union[str, os.PathLike[str], Box]
        The source of the CZI image data, which can be a file path or a Box object.
    model : List[str]
        A list to store the model names of the detectors.
    name : List[str]
        A list to store the names of the detectors.
    Id : List[str]
        A list to store the IDs of the detectors.
    modeltype : List[str]
        A list to store the types of the detectors.
    gain : List[float]
        A list to store the gain values of the detectors.
    zoom : List[float]
        A list to store the zoom values of the detectors.
    amplificationgain : List[float]
        A list to store the amplification gain values of the detectors.
    verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods
    -------
    __post_init__():
        Initializes the detector information by reading from the CZI image data.
    """

    czisource: Union[str, os.PathLike[str], Box]
    model: List[str] = field(init=False, default_factory=lambda: [])
    name: List[str] = field(init=False, default_factory=lambda: [])
    Id: List[str] = field(init=False, default_factory=lambda: [])
    modeltype: List[str] = field(init=False, default_factory=lambda: [])
    gain: List[float] = field(init=False, default_factory=lambda: [])
    zoom: List[float] = field(init=False, default_factory=lambda: [])
    amplificationgain: List[float] = field(init=False, default_factory=lambda: [])
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading Detector Information from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        # check if there are any detector entries inside the dictionary
        if czi_box.ImageDocument.Metadata.Information.Instrument is not None:
            # get the data for the detectors
            detectors = (
                czi_box.ImageDocument.Metadata.Information.Instrument.Detectors.Detector
            )

            # check for detector Id, Name, Model and Type
            if isinstance(detectors, Box):
                self.Id.append(detectors.Id)
                self.name.append(detectors.Name)
                self.model.append(detectors.Model)
                self.modeltype.append(detectors.Type)
                self.gain.append(detectors.Gain)
                self.zoom.append(detectors.Zoom)
                self.amplificationgain.append(detectors.AmplificationGain)

            # and do this differently in case of a list of detectors
            elif isinstance(detectors, BoxList):
                for d in range(len(detectors)):
                    self.Id.append(detectors[d].Id)
                    self.name.append(detectors[d].Name)
                    self.model.append(detectors[d].Model)
                    self.modeltype.append(detectors[d].Type)
                    self.gain.append(detectors[d].Gain)
                    self.zoom.append(detectors[d].Zoom)
                    self.amplificationgain.append(detectors[d].AmplificationGain)

        elif czi_box.ImageDocument.Metadata.Information.Instrument is None:
            if self.verbose:
                logger.info("No Detector(s) information found.")
            self.model = [None]
            self.name = [None]
            self.Id = [None]
            self.modeltype = [None]
            self.gain = [None]
            self.zoom = [None]
            self.amplificationgain = [None]



from typing import Union, Optional, Annotated, Dict
from dataclasses import dataclass, field
from box import Box, BoxList
import os
import numpy as np
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box
from czitools.metadata_tools.helper import ValueRange

logger = logging_tools.set_logging()


@dataclass
class CziScaling:
    """
    A class to handle scaling information from CZI image data.
    Attributes:
        czisource (Union[str, os.PathLike[str], Box]): The source of the CZI image data.
        X (Optional[float]): The scaling value for the X dimension in microns.
        Y (Optional[float]): The scaling value for the Y dimension in microns.
        Z (Optional[float]): The scaling value for the Z dimension in microns.
        X_sf (Optional[float]): The downscaled scaling value for the X dimension in microns.
        Y_sf (Optional[float]): The downscaled scaling value for the Y dimension in microns.
        ratio (Optional[Dict[str, float]]): The scaling ratios for XY, ZX, and ZX_sf.
        unit (Optional[str]): The unit of measurement for scaling, default is "micron".
        zoom (Annotated[float, ValueRange(0.01, 1.0)]): The zoom factor, default is 1.0.
        verbose (bool): Flag to enable verbose logging.
    Methods:
        __post_init__(): Initializes the scaling values from the CZI image data.
        _safe_get_scale(dist: BoxList, idx: int) -> Optional[float]: Safely retrieves the scaling value for a given dimension.
    """

    czisource: Union[str, os.PathLike[str], Box]
    X: Optional[float] = field(init=False, default=None)
    Y: Optional[float] = field(init=False, default=None)
    Z: Optional[float] = field(init=False, default=None)
    X_sf: Optional[float] = field(init=False, default=None)
    Y_sf: Optional[float] = field(init=False, default=None)
    ratio: Optional[Dict[str, float]] = field(init=False, default=None)
    # ratio_sf: Optional[Dict[str, float]] = field(init=False, default=None)
    # scalefactorXY: Optional[float] = field(init=False, default=None)
    unit: Optional[str] = field(init=True, default="micron")
    zoom: Annotated[float, ValueRange(0.01, 1.0)] = field(init=True, default=1.0)
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading Scaling from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        if czi_box.has_scale:
            distances = czi_box.ImageDocument.Metadata.Scaling.Items.Distance

            # get the scaling values for X,Y and Z
            self.X = np.round(
                self._safe_get_scale(distances, 0, verbose=self.verbose), 3
            )
            self.Y = np.round(
                self._safe_get_scale(distances, 1, verbose=self.verbose), 3
            )
            self.Z = np.round(
                self._safe_get_scale(distances, 2, verbose=self.verbose), 3
            )

            # calc the scaling values for X,Y when applying downscaling
            self.X_sf = np.round(self.X * (1 / self.zoom), 3)
            self.Y_sf = np.round(self.Y * (1 / self.zoom), 3)

            # calc the scaling ratio
            self.ratio = {
                "xy": np.round(self.X / self.Y, 3),
                "zx": np.round(self.Z / self.X, 3),
                "zx_sf": np.round(self.Z / self.X_sf, 3),
            }

        elif not czi_box.has_scale:
            if self.verbose:
                logger.warning("No scaling information found.")

    @staticmethod
    def _safe_get_scale(
        dist: BoxList, idx: int, verbose: bool = False
    ) -> Optional[float]:
        scales = ["X", "Y", "Z"]

        try:
            # get the scaling value in [micron]
            sc = float(dist[idx].Value) * 1000000

            # check for the value = 0.0
            if sc == 0.0:
                sc = 1.0
                if verbose:
                    logger.info(
                        "Detected Scaling = 0.0 for "
                        + scales[idx]
                        + " Using default = 1.0 [micron]."
                    )
            return sc

        except (IndexError, TypeError, AttributeError):
            if verbose:
                logger.info(
                    "No "
                    + scales[idx]
                    + "-Scaling found. Using default = 1.0 [micron]."
                )
            return 1.0



from dataclasses import dataclass, field
from enum import Enum


@dataclass
class ValueRange:
    lo: float
    hi: float


class AttachmentType(Enum):
    SlidePreview = 1
    Label = 2
    Prescan = 3


class DictObj:
    """
    This class recursively converts a dictionary into an object, allowing attribute-style access to dictionary keys.
    See https://joelmccune.com/python-dictionary-as-object/
    Attributes:
        in_dict (dict): The dictionary to be converted into an object.
    Methods:
        __init__(in_dict: dict) -> None:
            Initializes the DictObj instance by converting the input dictionary into an object.
    """

    # TODO: is this class still needed because of using python-box

    def __init__(self, in_dict: dict) -> None:
        assert isinstance(in_dict, dict)

        for key, val in in_dict.items():
            if isinstance(val, (list, tuple)):
                setattr(
                    self, key, [DictObj(x) if isinstance(x, dict) else x for x in val]
                )
            else:
                setattr(self, key, DictObj(val) if isinstance(val, dict) else val)



from typing import Union, Optional
from dataclasses import dataclass, field
from box import Box
import os
from czitools.utils import logging_tools
from czitools.utils.box import get_czimd_box

logger = logging_tools.set_logging()


@dataclass
class CziMicroscope:
    """
    A class to represent a microscope from CZI image data.
    Attributes:
    ----------
    czisource : Union[str, os.PathLike[str], Box]
        The source of the CZI image data, which can be a file path or a Box object.
    Id : Optional[str]
        The identifier of the microscope. Initialized in __post_init__.
    Name : Optional[str]
        The name of the microscope. Initialized in __post_init__.
    System : Optional[str]
        The system of the microscope. Initialized in __post_init__.
    verbose (bool): Flag to enable verbose logging. Initialized to False.
    Methods:
    -------
    __post_init__():
        Initializes the microscope information from the CZI image data.
    """

    czisource: Union[str, os.PathLike[str], Box]
    Id: Optional[str] = field(init=False)
    Name: Optional[str] = field(init=False)
    System: Optional[str] = field(init=False)
    verbose: bool = False

    def __post_init__(self):
        if self.verbose:
            logger.info("Reading Microscope Information from CZI image data.")

        if isinstance(self.czisource, Box):
            czi_box = self.czisource
        else:
            czi_box = get_czimd_box(self.czisource)

        if czi_box.ImageDocument.Metadata.Information.Instrument is None:
            self.Id = None
            self.Name = None
            self.System = None
            logger.info("No Microscope information found.")

        else:
            self.Id = (
                czi_box.ImageDocument.Metadata.Information.Instrument.Microscopes.Microscope.Id
            )
            self.Name = (
                czi_box.ImageDocument.Metadata.Information.Instrument.Microscopes.Microscope.Name
            )
            self.System = (
                czi_box.ImageDocument.Metadata.Information.Instrument.Microscopes.Microscope.System
            )



# -*- coding: utf-8 -*-

#################################################################
# File        : pixels.py
# Author      : sebi06
#
# Disclaimer: The code is purely experimental. Feel free to
# use it at your own risk.
#
#################################################################

from pylibCZIrw import czi as pyczi
import numpy as np
from typing import List, Dict, Tuple, Optional, Union


def check_if_rgb(pixeltypes: Dict) -> Tuple[Dict[int, bool], bool]:
    """
    Check if the pixel types are RGB and if they are consistent.
    Args:
        pixeltypes (Dict): A dictionary where keys are pixel identifiers and values are pixel type strings.
    Returns:
        Tuple[Dict, bool]: A tuple containing:
            - A dictionary where keys are pixel identifiers and values are booleans indicating if the pixel type is RGB.
            - A boolean flag indicating if all pixel types in the input dictionary are the same.
    """

    is_rgb = {}

    for k, v in pixeltypes.items():
        if "Bgr" in v:
            is_rgb[k] = True
        else:
            is_rgb[k] = False

    # flag to check if all elements are same
    is_consistent = True

    # extracting value to compare
    test_val = list(pixeltypes.values())[0]

    is_consistent = all(value == test_val for value in pixeltypes.values())

    return is_rgb, is_consistent


def get_dtype_fromstring(
    pixeltype: str,
) -> Tuple[Optional[np.dtype], Optional[int]]:
    """
    Determine the numpy data type and maximum value based on the given pixel type string.
    Parameters:
    pixeltype (str): The pixel type as a string. Possible values include "gray16", "Gray16",
                     "gray8", "Gray8", "bgr48", "Bgr48", "bgr24", "Bgr24", "bgr96float", "Bgr96Float".
    Returns:
    Tuple[Optional[np.dtype], Optional[int]]: A tuple containing the numpy data type and the maximum value
                                              for the given pixel type. If the pixel type is not recognized,
                                              both elements of the tuple will be None.
    """
    dtype = None
    maxvalue = None

    if pixeltype == "gray16" or pixeltype == "Gray16":
        dtype = np.dtype(np.uint16)
        maxvalue = 65535

    if pixeltype == "gray8" or pixeltype == "Gray8":
        dtype = np.dtype(np.uint8)
        maxvalue = 255

    if pixeltype == "bgr48" or pixeltype == "Bgr48":
        dtype = np.dtype(np.uint16)
        maxvalue = 65535

    if pixeltype == "bgr24" or pixeltype == "Bgr24":
        dtype = np.dtype(np.uint8)
        maxvalue = 255

    if pixeltype == "bgr96float" or pixeltype == "Bgr96Float":
        dtype = np.dtype(np.uint16)
        maxvalue = 65535

    return dtype, maxvalue


def check_scenes_shape(czidoc: pyczi.CziReader, size_s: Union[int, None]) -> bool:
    """Check if all scenes have the same shape.

    Args:
        czidoc (pyczi.CziReader): CziReader to read the properties
        size_s (Union[int, None]): Size of scene dimension

    Returns:
        bool: True is all scenes have identical XY shape
    """
    scene_width = []
    scene_height = []
    scene_shape_is_consistent = False

    if size_s is not None:
        for s in range(size_s):
            scene_width.append(czidoc.scenes_bounding_rectangle[s].w)
            scene_height.append(czidoc.scenes_bounding_rectangle[s].h)

        # check if all entries in list are the same
        sw = scene_width.count(scene_width[0]) == len(scene_width)
        sh = scene_height.count(scene_height[0]) == len(scene_height)

        # only if entries for X and Y are all the same as the shape is consistent
        if sw is True and sh is True:
            scene_shape_is_consistent = True

    else:
        scene_shape_is_consistent = True

    return scene_shape_is_consistent


def get_dimorder(dim_string: str) -> Tuple[Dict, List, int]:
    """
    Extracts the order and indices of dimensions from a given dimension string.
    Args:
        dim_string (str): A string representing the dimensions.
    Returns:
        Tuple[Dict, List, int]: A tuple containing:
            - A dictionary with dimensions as keys and their indices in the string as values.
            - A list of indices corresponding to the dimensions in the order they appear in the predefined list.
            - An integer representing the number of valid dimensions found in the string.
    """

    dimindex_list = []
    dims = ["R", "I", "M", "H", "V", "B", "S", "T", "C", "Z", "Y", "X", "A"]
    dims_dict = {}

    # loop over all dimensions and find the index
    for d in dims:
        dims_dict[d] = dim_string.find(d)
        dimindex_list.append(dim_string.find(d))

    # check if a dimension really exists
    numvalid_dims = sum(i >= 0 for i in dimindex_list)

    return dims_dict, dimindex_list, numvalid_dims
